from pathlib import Path
from typing import Any, List, Mapping, Optional, Union
from enum import Enum
import re

from pydantic import BaseModel, Field, FilePath, validator, Extra

from ..core.errors import Exc, Ok, Err
from ..database import Database
from .sql import SqlTask


class Destination(BaseModel):
    supports_schemas: bool
    db_type: str
    db: Optional[str]
    tmp_schema: Optional[str]
    db_schema: Optional[str] = Field(None, alias="schema")
    table: str

    class Config:
        extra = Extra.forbid

    @validator("tmp_schema")
    def can_use_tmp_schema(cls, v, values):
        if v is not None and not values["supports_schemas"]:
            raise ValueError(
                f'tmp_schema not supported for database of type {values["db_type"]}'
            )

        return v

    @validator("db_schema")
    def can_use_schema(cls, v, values):
        if v is not None and not values["supports_schemas"]:
            raise ValueError(
                f'schema not supported for database of type {values["db_type"]}'
            )

        return v


class Config(BaseModel):
    sql_folder: Path
    file_name: FilePath
    delete_key: Optional[str]
    materialisation: str
    destination: Destination
    columns: Optional[List[Union[str, Mapping[str, Any]]]] = list()
    table_properties: Mapping[str, Any] = dict()
    post_hook: List[Mapping[str, Any]] = list()

    class Config:
        extra = Extra.forbid

    @validator("file_name", pre=True)
    def file_name_plus_folder(cls, v, values):
        return Path(values["sql_folder"], v)

    @validator("materialisation")
    def incremental_has_delete_key(cls, v, values):
        if v not in ("table", "view", "incremental"):
            raise ValueError(f'"{v}". Valid materialisations: table, view, incremental')
        elif v != "incremental" and values.get("delete_key") is not None:
            raise ValueError('"delete_key" is invalid in non-incremental loads')
        elif v == "incremental" and values.get("delete_key") is None:
            raise ValueError('"delete_key" is required for incremental loads')
        else:
            return v


class OnFailValue(str, Enum):
    skip = "skip"
    no_skip = "no_skip"


class CompileConfig(BaseModel):
    supports_schemas: bool
    db_type: str

    delete_key: Optional[str]
    materialisation: Optional[str]
    tmp_schema: Optional[str]
    db_schema: Optional[str] = Field(None, alias="schema")
    table: Optional[str]
    columns: Optional[List[Union[str, Mapping[str, Any]]]] = list()
    table_properties: Mapping[str, Any] = dict()
    post_hook: List[Mapping[str, Any]] = list()
    tags: Optional[List[str]]
    parents: Optional[List[str]]
    on_fail: Optional[OnFailValue]

    class Config:
        extra = Extra.forbid

    @validator("materialisation")
    def incremental_has_delete_key(cls, v, values):
        if v not in ("table", "view", "incremental"):
            raise ValueError(f'"{v}". Valid materialisations: table, view, incremental')
        elif v != "incremental" and values.get("delete_key") is not None:
            raise ValueError('"delete_key" is invalid in non-incremental loads')
        elif v == "incremental" and values.get("delete_key") is None:
            raise ValueError('"delete_key" is required for incremental loads')
        else:
            return v

    @validator("tmp_schema")
    def can_use_tmp_schema(cls, v, values):
        if v is not None and not values["supports_schemas"]:
            raise ValueError(
                f'tmp_schema not supported for database of type {values["db_type"]}'
            )

        return v

    @validator("db_schema")
    def can_use_schema(cls, v, values):
        if v is not None and not values["supports_schemas"]:
            raise ValueError(
                f'schema not supported for database of type {values["db_type"]}'
            )

        return v


class AutoSqlTask(SqlTask):
    def config(self, **config):
        if "task_name" in self._config_input:
            del self._config_input["task_name"]

        # if "columns" in config:
        #     self._has_tests = True

        conn_names_list = [
            n for n, c in self.connections.items() if isinstance(c, Database)
        ]

        # set the target db for execution
        # this check needs to happen here so we can pass db_features and db_type to the validator
        if (
            isinstance(config.get("destination"), dict)
            and config["destination"].get("db") is not None
        ):
            if config["destination"]["db"] not in conn_names_list:
                return Err(
                    "task_definition",
                    "destination_db_not_in_settings",
                    db=config["destination"]["db"],
                )
            self._target_db = config["destination"]["db"]
        else:
            self._target_db = self._default_db

        if isinstance(config.get("destination"), dict):
            config["destination"].update(
                {
                    "supports_schemas": not self.target_db.feature("NO SCHEMA SUPPORT"),
                    "db_type": self.target_db.db_type,
                }
            )

        try:
            self.task_config = Config(
                sql_folder=self.run_arguments["folders"]["sql"], **config
            )
        except Exception as e:
            return Exc(e)

        # We compile first to allow changes to the config
        # Template compilation
        self.compiler.update_globals(
            src=lambda x: self.src(x, connection=self._target_db),
            config=self.config_macro,
        )

        self.allow_config = True

        self.prepared_sql_query = self.compiler.prepare(self.task_config.file_name)
        self.sql_query = self.prepared_sql_query.compile()

        # After the first compilation, config shouldn't be executed again
        self.allow_config = False

        # Output calculation
        db_schema = self.task_config.destination.db_schema

        if self.task_config.destination.tmp_schema is not None:
            tmp_db_schema = self.task_config.destination.tmp_schema
        else:
            tmp_db_schema = db_schema

        base_table_name = self.task_config.destination.table

        if db_schema is None:
            self.schema = None
            self.table = self.out(base_table_name, self.target_db)
        else:
            obj = self.out(f"{db_schema}.{base_table_name}", self.target_db)
            self.schema = obj.split(".")[0]
            self.table = obj.split(".")[1]

        if tmp_db_schema is None:
            self.tmp_schema = None
            self.tmp_table = self.out(f"sayn_tmp_{base_table_name}", self.target_db)
        else:
            obj = self.out(
                f"{tmp_db_schema}.sayn_tmp_{base_table_name}", self.target_db
            )
            self.tmp_schema = obj.split(".")[0]
            self.tmp_table = obj.split(".")[1]

        self.materialisation = self.task_config.materialisation
        self.delete_key = self.task_config.delete_key

        # DDL validation
        result = self.target_db._validate_ddl(
            self.task_config.columns,
            self.task_config.table_properties,
            self.task_config.post_hook,
        )
        if result.is_err:
            return result
        else:
            self.ddl = result.value

        if self.run_arguments["command"] == "test" and len(self.ddl["columns"]) != 0:
            result = self.target_db._construct_tests(
                self.ddl["columns"], self.table, self.schema
            )
            if result.is_err:
                return result
            else:
                self.test_query = result.value[0]
                self.test_breakdown = result.value[1]

            if self.test_query is not None:
                self._has_tests = True

        return Ok()

    def config_macro(self, **config):
        if self.allow_config:
            config.update(
                {
                    "supports_schemas": not self.target_db.feature("NO SCHEMA SUPPORT"),
                    "db_type": self.target_db.db_type,
                }
            )
            task_config_override = CompileConfig(**config)

            self.task_config.materialisation = (
                task_config_override.materialisation or self.task_config.materialisation
            )
            self.task_config.destination.db_schema = (
                task_config_override.db_schema or self.task_config.destination.db_schema
            )
            self.task_config.destination.tmp_schema = (
                task_config_override.tmp_schema
                or self.task_config.destination.tmp_schema
            )
            self.task_config.destination.table = (
                task_config_override.table or self.task_config.destination.table
            )
            self.task_config.columns = (
                task_config_override.columns or self.task_config.columns
            )
            self.task_config.table_properties = (
                task_config_override.table_properties
                or self.task_config.table_properties
            )
            self.task_config.post_hook = (
                task_config_override.post_hook or self.task_config.post_hook
            )
            self.task_config.delete_key = (
                task_config_override.delete_key or self.task_config.delete_key
            )
            # Sent to the wrapper
            if task_config_override.on_fail is not None:
                self._config_input["on_fail"] = task_config_override.on_fail

            if task_config_override.tags is not None:
                self._config_input["tags"] = task_config_override.tags

            if task_config_override.parents is not None:
                self._config_input["parents"] = task_config_override.parents

        # Returns an empty string to avoid productin incorrect sql
        return ""

    def setup(self):
        if self.needs_recompile:
            self.sql_query = self.prepared_sql_query.compile()

            if self._has_tests:
                schema = self.task_config.destination.db_schema
                table = self.task_config.destination.table
                obj = self.src(f"{schema}.{table}", self.target_db)
                test_schema = obj.split(".")[0]
                test_table = obj.split(".")[1]
                result = self.target_db._construct_tests(
                    self.ddl["columns"], test_table, test_schema
                )
                if result.is_err:
                    return result
                else:
                    self.test_query = result.value[0]
                    self.test_breakdown = result.value[1]

        return Ok()

    def execute(self, execute, debug):
        if self.run_arguments["debug"]:
            self.write_compilation_output(self.sql_query, "select")
        else:
            self.write_compilation_output(self.sql_query)

        if self.materialisation == "view":
            # View
            step_queries = self.target_db.replace_view(
                self.table,
                self.sql_query,
                schema=self.schema,
                **self.ddl,
            )

        elif (
            self.materialisation == "table"
            or self.run_arguments["full_load"]
            or self.target_db._requested_objects[""][self.schema or ""][self.table].get(
                "type"
            )
            is None
        ):
            # Full load or target table missing
            if self.target_db.feature("CANNOT CHANGE SCHEMA"):
                # Use destination schema if the db doesn't support schema changes
                tmp_schema = self.schema
            else:
                tmp_schema = self.tmp_schema

            step_queries = self.target_db.replace_table(
                self.table,
                self.sql_query,
                schema=self.schema,
                tmp_schema=tmp_schema,
                **self.ddl,
            )

        else:
            # Incremental load
            step_queries = self.target_db.merge_query(
                self.table,
                self.sql_query,
                self.delete_key,
                schema=self.schema,
                tmp_schema=self.tmp_schema,
                **self.ddl,
            )

        self.set_run_steps(list(step_queries.keys()))

        for step, query in step_queries.items():
            with self.step(step):
                if debug:
                    self.write_compilation_output(query, step.replace(" ", "_").lower())
                if execute:
                    try:
                        self.target_db.execute(query)
                    except Exception as e:
                        return Exc(e)

        return Ok()

    def compile(self):
        return self.execute(False, self.run_arguments["debug"])

    def run(self):
        return self.execute(True, self.run_arguments["debug"])

    def test(self):
        step_queries = {
            "Write Test Query": self.test_query,
            "Execute Test Query": self.test_query,
        }
        breakdown = self.get_test_breakdown(self.test_breakdown)

        if self.test_query is None:
            self.info("Nothing to be done")
            return self.success()
        else:
            self.set_run_steps(list(step_queries.keys()))

            for step, query in step_queries.items():
                with self.step(step):
                    if "Write" in step:
                        self.write_compilation_output(query, "test")
                    if "Execute" in step:
                        try:
                            result = self.default_db.read_data(query)
                        except Exception as e:
                            return Exc(e)

            if len(result) == 0:
                return self.test_sucessful(breakdown)
            else:
                errout, failed = self.test_failure(
                    breakdown, result, self.run_arguments["debug"]
                )
                problematic_values_query = self.default_db.test_problematic_values(
                    failed, self.table, self.schema
                )

                for query in problematic_values_query.split(";"):
                    if query.strip():
                        header = re.search(r"--.*?--", query).group(0)
                        self.info("")
                        self.info(header)
                        self.info(
                            "===================================================================="
                        )
                        self.info(
                            re.sub(r"--.*?--", "", query).replace("\n", " ").strip()
                            + ";"
                        )
                        self.info(
                            "===================================================================="
                        )
                        self.info("")

                self.write_compilation_output(
                    problematic_values_query, "test_problematic_values"
                )

                return errout



from .task import Task


class DummyTask(Task):
    def config(self):
        return self.success()

    def setup(self):
        return self.success()

    def compile(self):
        return self.success()

    def run(self):
        return self.success()

    def test(self):
        return self.success()



from collections import Counter
import csv
from pathlib import Path
import tempfile
from typing import List, Optional, Union

from pydantic import BaseModel, constr, validator, Extra
from sqlalchemy import create_engine

from ..core.errors import DBError
from . import Database, Columns, Hook, BaseDDL


DistributionStr = constr(regex=r"even|all|key([^,]+)")


class DDL(BaseDDL):
    class Properties(BaseModel):
        class Sorting(BaseModel):
            type: Optional[str]
            columns: List[str]

            class Config:
                extra = Extra.forbid

            @validator("type")
            def validate_type(cls, v):
                if v.upper() not in ("COMPOUND", "INTERLEAVED"):
                    raise ValueError(
                        'Sorting type must be one of "COMPOUND" or "INTERLEAVED"'
                    )

                return v.upper()

        distribution: Optional[DistributionStr]
        sorting: Optional[Sorting]

        class Config:
            extra = Extra.forbid

    columns: List[Union[str, Columns]] = list()
    properties: Optional[Properties]
    post_hook: List[Hook] = list()

    class Config:
        extra = Extra.forbid

    @validator("columns", pre=True)
    def transform_str_cols(cls, v):
        if v is not None and isinstance(v, List):
            return [{"name": c} if isinstance(c, str) else c for c in v]
        else:
            return v

    @validator("columns")
    def columns_unique(cls, v):
        dupes = {k for k, v in Counter([e.name for e in v]).items() if v > 1}
        if len(dupes) > 0:
            raise ValueError(f"Duplicate columns: {','.join(dupes)}")
        else:
            return v

    @validator("properties")
    def validate_distribution(cls, v):
        if v is not None:
            if v.distribution is not None:
                distribution = {
                    "type": v.distribution.upper()
                    if v.distribution.lower() in ("even", "all")
                    else "KEY"
                }
                if distribution["type"] == "KEY":
                    distribution["column"] = v.distribution[len("key") + 1 : -1]
                v.distribution = distribution

            return v
        else:
            return v

    def get_ddl(self):
        result = self.base_ddl()
        properties = list()
        if self.properties is not None:
            if self.properties.distribution is not None:
                properties.append({"distribution": self.properties.distribution})
                result["distribution"] = self.properties.distribution

            if self.properties.sorting is not None:
                properties.append({"sorting": self.properties.sorting})
                result["sorting"] = self.properties.sorting

        return result


class Redshift(Database):
    DDL = DDL
    _boto_session = None
    bucket = None
    access_key_id = None
    secret_access_key = None
    session_token = None
    profile = None

    def feature(self, feature):
        return feature in (
            "NEEDS CASCADE",
            "CANNOT CHANGE SCHEMA",
            "CANNOT SPECIFY DDL IN SELECT",
        )

    def create_engine(self, settings):
        # Create engine using the connect_args argument to create_engine
        if "connect_args" not in settings:
            settings["connect_args"] = dict()

        region = settings.pop("region", None)

        if "cluster_id" in settings and "password" not in settings:
            # if a cluster_id is given and no password, we use boto to
            # complete the credentials
            settings["connect_args"]["iam"] = True
            settings["connect_args"]["db_user"] = settings.pop("user")

            settings["connect_args"]["cluster_identifier"] = settings.pop("cluster_id")

            if "aws_access_key_id" in settings or "aws_secret_access_key" in settings:
                self.access_key_id = settings.pop("aws_access_key_id", None)
                self.secret_access_key = settings.pop("aws_secret_access_key", None)
                self.session_token = settings.pop("aws_session_token", None)

                if self.access_key_id is None or self.secret_access_key is None:
                    raise DBError(
                        self.name,
                        self.db_type,
                        "Error retrieving AWS access key credentials",
                    )
            elif "profile" in settings:
                self.profile = settings.pop("profile")

        if "bucket" in settings:
            self.bucket = {"name": None, "region": None, "role": None}
            bucket = settings.pop("bucket")
            if isinstance(bucket, str):
                self.bucket["name"] = bucket
            elif isinstance(bucket, dict):
                if "name" in bucket:
                    self.bucket["name"] = bucket["name"]

                if "region" in bucket:
                    self.bucket["region"] = bucket["region"]

                if "role" in bucket:
                    self.bucket["role"] = bucket["role"]

            # Setup the boto3 session
            import boto3

            self._boto_session = boto3.Session(
                profile_name=self.profile,
                region_name=region,
                aws_access_key_id=self.access_key_id,
                aws_secret_access_key=self.secret_access_key,
                aws_session_token=self.session_token,
            )

        dbname = settings.pop("dbname")

        db_parameters = [
            "host",
            "user",
            "password",
            "port",
        ]

        for param in db_parameters:
            if param in settings:
                settings["connect_args"][param] = settings.pop(param)

        if self.access_key_id is not None:
            settings["connect_args"]["access_key_id"] = self.access_key_id

        if self.secret_access_key is not None:
            settings["connect_args"]["secret_access_key"] = self.secret_access_key

        if self.session_token is not None:
            settings["connect_args"]["session_token"] = self.session_token

        if self.profile is not None:
            settings["connect_args"]["profile"] = self.profile

        return create_engine(f"redshift+redshift_connector:///{dbname}", **settings)

    def execute(self, script):
        conn = self.engine.raw_connection()
        with conn.cursor() as cursor:
            for s in script.split(";"):
                if len(s.strip()) > 0:
                    cursor.execute(s)
                    cursor.execute("COMMIT")

    def _list_databases(self):
        report = self.read_data("SELECT datname FROM pg_database")
        dbs = [re["datname"] for re in report]
        return dbs

    def _load_data_batch(self, table, data, schema, db):
        """Implements the load of a single data batch for `load_data`.

        Defaults to an insert many statement, but it's overloaded for specific
        database connector for more efficient methods.

        Args:
            table (str): The name of the target table
            data (list): A list of dictionaries to load
            schema (str): An optional schema to reference the table
        """
        # if no bucket is supplied, the old _load_data_batch function is used
        if self.bucket is None:
            return super(Database, self)._load_data_batch(table, data, schema, db)

        full_table_name = f"{'' if db is None else db + '.'}{'' if schema is None else schema + '.'}{table}"
        template = self._jinja_env.get_template("redshift_load_batch.sql")
        fname = "batch.csv"

        s3_client = self._boto_session.client("s3")

        with tempfile.TemporaryDirectory() as tmpdirname:
            with (Path(tmpdirname) / fname).open("w") as f:
                writer = csv.DictWriter(
                    f, fieldnames=data[0].keys(), delimiter="|", escapechar="\\"
                )

                writer.writerows(data)

            with (Path(tmpdirname) / fname).open("rb") as f:
                s3_client.upload_fileobj(f, self.bucket, fname)

            self.execute(
                template.render(
                    full_table_name=full_table_name,
                    temp_file_directory=tmpdirname,
                    temp_file_name=fname,
                    bucket=self.bucket,
                    region=self.bucket_region,
                )
            )

    def merge_tables(
        self,
        src_table,
        dst_table,
        delete_key,
        cleanup=True,
        src_schema=None,
        dst_schema=None,
        src_db=None,
        dst_db=None,
        **ddl,
    ):
        src_table = fully_qualify(src_table, src_schema, src_db)
        dst_table = fully_qualify(dst_table, dst_schema, dst_db)

        template = self._jinja_env.get_template("redshift_merge_tables.sql")
        return template.render(
            dst_table=dst_table,
            src_table=src_table,
            cleanup=True,
            delete_key=delete_key,
        )


def fully_qualify(name, schema=None, db=None):
    return f"{db+'.' if db is not None else ''}{schema+'.' if schema is not None else ''}{name}"



import csv
import io
from sqlalchemy import create_engine

from . import Database

db_parameters = ["host", "user", "password", "port", "dbname"]


class Postgresql(Database):
    def feature(self, feature):
        return feature in (
            "NEEDS CASCADE",
            "CAN REPLACE VIEW",
            "CANNOT SPECIFY DDL IN SELECT",
        )

    def create_engine(self, settings):
        # Create engine using the connect_args argument to create_engine
        if "connect_args" not in settings:
            settings["connect_args"] = dict()
        for param in db_parameters:
            if param in settings:
                settings["connect_args"][param] = settings.pop(param)

        return create_engine("postgresql://", **settings)

    def _list_databases(self):
        report = self.read_data("SELECT datname FROM pg_database;")
        dbs = [re["datname"] for re in report]
        return dbs

    def _load_data_batch(self, table, data, schema, db):
        full_table_name = f"{'' if db is None else db + '.'}{'' if schema is None else schema + '.'}{table}"
        copy_sql = f"COPY {full_table_name} FROM STDIN " "CSV DELIMITER ',' QUOTE '\"'"

        buffer = io.StringIO()
        writer = csv.DictWriter(buffer, fieldnames=data[0].keys())
        writer.writerows(data)
        buffer.seek(0)
        connection = self.engine.connect().connection
        with connection.cursor() as cursor:
            cursor.copy_expert(copy_sql, buffer)
            connection.commit()



from pathlib import Path
import logging

from .logger import Logger
from .log_formatter import LogFormatter


class FileLogger(Logger):
    fmt = LogFormatter(use_colour=False, output_ts=False)
    logger = None

    def __init__(self, folder, format=None):
        if format is None:
            format = "%(asctime)s|%(levelname)s|%(message)s"

        formatter = logging.Formatter(format)

        log_file = Path(folder, "sayn.log")
        if not log_file.parent.exists():
            log_file.parent.mkdir(parents=True)

        handler = logging.FileHandler(log_file)
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)

        logger = logging.getLogger(__name__)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)

        self.logger = logger

    def print(self, s=None):
        if s is not None:
            if s["level"] == "info":
                func = self.logger.info
            elif s["level"] == "error":
                func = self.logger.error
            elif s["level"] == "warning":
                func = self.logger.warning
            else:
                func = self.logger.debug
            s = s["message"]

            if isinstance(s, str):
                s = [s]
            elif not isinstance(s, list):
                raise ValueError("error in logging print")

            func(f"{s[0]}")
            for e in s[1:]:
                for l in e.split("\n"):
                    func(f"{l}")



from datetime import date, datetime, timedelta
import traceback

from colorama import init, Fore, Style

from ..utils.misc import group_list

init(autoreset=True)


def human(obj):
    if isinstance(obj, timedelta):
        secs = obj.total_seconds()
        mins = secs / 60.0
        hours = mins / 60.0

        if hours > 1.0:
            r = str(round(hours, 1))
            return f"{r if r[-1] != '0' else r[:-2]}h"

        if mins > 1.0:
            r = str(round(mins, 1))
            return f"{r if r[-1] != '0' else r[:-2]}m"

        elif secs > 1.0:
            r = str(round(secs, 1))
            return f"{r if r[-1] != '0' else r[:-2]}s"

        else:
            msecs = secs * 1000.0
            r = str(round(msecs, 1))
            return f"{r if r[-1] != '0' else r[:-2]}ms"
    elif isinstance(obj, datetime):
        return obj.strftime("%H:%M")
    elif isinstance(obj, str):
        return obj.replace("_", " ").title()
    else:
        return obj


class LogFormatter:
    use_colour = True
    use_icons = True
    output_ts = False

    def __init__(self, use_colour=True, use_icons=True, output_ts=False):
        self.use_colour = use_colour
        self.output_ts = output_ts
        self.use_icons = True

    # Styling methods

    def dim(self, s):
        if self.use_colour:
            return f"{Style.DIM}{s}{Style.NORMAL}"
        else:
            return s

    def bright(self, s):
        if self.use_colour:
            return f"{Style.BRIGHT}{s}{Style.NORMAL}"
        else:
            return s

    def red(self, s):
        if self.use_colour:
            return f"{Fore.RED}{s}{Fore.RESET}"
        else:
            return s

    def bright_red(self, s):
        if self.use_colour:
            return f"{Fore.RED}{Style.BRIGHT}{s}{Style.NORMAL}{Fore.RESET}"
        else:
            return s

    def yellow(self, s):
        if self.use_colour:
            return f"{Fore.YELLOW}{s}{Fore.RESET}"
        else:
            return s

    def bright_yellow(self, s):
        if self.use_colour:
            return f"{Fore.YELLOW}{Style.BRIGHT}{s}{Style.NORMAL}{Fore.RESET}"
        else:
            return s

    def green(self, s):
        if self.use_colour:
            return f"{Fore.GREEN}{s}{Fore.RESET}"
        else:
            return s

    def bright_green(self, s):
        if self.use_colour:
            return f"{Fore.GREEN}{Style.BRIGHT}{s}{Style.NORMAL}{Fore.RESET}"
        else:
            return s

    def indent(self, s, l):
        if self.use_colour:
            return " " * l + s
        else:
            return s

    def good(self, s):
        if self.use_colour and self.use_icons:
            return self.green(f"✔ {s}")
        else:
            return s

    def info(self, s):
        if self.use_colour and self.use_icons:
            return f"ℹ {s}"
        else:
            return s

    def warn(self, s):
        if self.use_colour and self.use_icons:
            return self.yellow(f"⚠ {s}")
        else:
            return s

    def bad(self, s):
        if self.use_colour and self.use_icons:
            return self.red(f"✖ {s}")
        else:
            return s

    def join(self, l, sep, f):
        if self.use_colour:
            return sep.join(f(s) for s in l)
        else:
            return sep.join(l)

    def blist(self, l):
        if self.use_colour:
            return self.join(l, ", ", self.bright)
        else:
            return ", ".join(l)

    # Event handling methods

    def unhandled(self, event, context, stage, details):
        ignored = (
            "project_git_commit",
            "sayn_version",
            "project_name",
            "ts",
            "run_id",
            "task",
            "task_order",
            "total_tasks",
        )
        ctx = details["task"] if context == "task" else context
        return {
            "level": "error",
            "message": f"Unhandled: {ctx}::{stage}::{event}: "
            + {k: v for k, v in details.items() if k not in ignored}.__str__(),
        }

    def message(self, level, message, details):
        if not isinstance(message, list):
            message = [message]

        out = []
        ts = f"[{human(details['ts'])}]" if self.output_ts else ""

        for m in message:
            if level == "error":
                out.append(self.bad(f"{ts} {m}"))
            elif level == "warning":
                out.append(self.warn(f"{ts} {m}"))
            elif level == "debug":
                out.append(self.dim(f"{ts} {m}"))
            else:
                out.append(f"{ts} {m}")

        return {"level": level, "message": out}

    def error_result(self, duration, error):  # noqa: C901
        level = "error"
        message = self.bad(error.__str__())
        duration = human(duration)

        if "error_message" in error.details:
            message = self.bad(error.details["error_message"])

        elif error.kind == "exception":
            exc = error.details["exception"]
            message = [self.bad(f"Failed ({duration}) {exc}")]
            message.extend(
                [
                    self.red(l)
                    for it in traceback.format_exception(
                        exc, value=exc, tb=exc.__traceback__
                    )
                    for l in it.split("\n")
                ]
            )

        elif error.kind == "dag" and error.code == "cycle_error":
            level = "error"
            message = self.bad(
                f"A cycle was detected in the dag: {' > '.join(error.details['path'])}"
            )

        elif error.kind == "dag" and error.code == "missing_parents":
            level = "error"
            message = ["Some parents are missing from dag"] + [
                self.red(f"In task {self.bright(task)}: {', '.join(parents)}")
                for task, parents in error.details["missing"].items()
            ]

        elif error.kind == "dag" and error.code == "missing_sources":
            level = "error"
            message = self.bad(error.details["error_message"])

        elif error.kind == "dag" and error.code == "empty_dag":
            level = "error"
            message = self.bad("No tasks defined in this project")

        elif error.kind == "dag" and error.code == "empty_group":
            level = "error"
            message = self.bad(
                f'Group "{error.details["group"]}" contains no tasks. Please check the "file_name" property'
            )

        elif error.kind == "task_query" and error.code == "query_overlap":
            level = "error"
            message = self.bad(
                f"{error.details['overlap']} specified both as include and exclude"
            )

        elif error.kind == "task_query" and error.code == "incorrect_syntax":
            level = "error"
            message = self.bad(f'Incorrect filter syntax "{error.details["query"]}"')

        elif error.kind == "task_query" and error.code == "undefined_tag":
            level = "error"
            message = self.bad(
                f'Tag not found in the project: "{error.details["tag"]}"'
            )

        elif error.kind == "task_query" and error.code == "undefined_group":
            level = "error"
            message = self.bad(
                f'Group not found in the project: "{error.details["group"]}"'
            )

        elif error.kind == "task_query" and error.code == "undefined_task":
            level = "error"
            message = self.bad(
                f'Task not found in the project: "{error.details["task"]}"'
            )

        elif error.code == "wrong_credentials":
            level = "error"
            message = self.bad(
                f'Connections {self.bright(", ".join(error.details["credentials"]))} not defined in project.yaml'
            )

        elif error.code == "missing_credentials":
            level = "error"
            message = self.bad(
                f'Connections {self.bright(", ".join(error.details["credentials"]))} are required by project.yaml'
            )

        elif error.code == "missing_credential_type":
            level = "error"
            message = self.bad(
                f'Connections {self.bright(", ".join(error.details["credentials"]))} have no type'
            )

        elif error.kind == "tasks" and error.code == "task_fail":
            level = "error"
            message = self.bad(error.details["message"])

        elif error.code == "parent_errors":
            level = "warning"
            parents = ", ".join(
                [f"{p} ({s.value})" for p, s in error.details["failed_parents"].items()]
            )
            message = self.warn(
                f"Skipping due to ancestors errors: {parents} ({duration})"
            )

        elif error.code == "setup_error":
            if error.details["status"].value == "skipped":
                level = "warning"
                message = self.warn(f"Skipping due to parent errors ({duration})")
            else:
                level = "error"
                message = self.bad(f"Failed during setup ({duration})")

        elif error.code == "validation_error":
            level = "error"
            message = [self.bad(f"Validation errors found ({duration})")]
            message.extend(
                [
                    self.red(
                        f"  In {' > '.join([str(item) for item in e['loc']])}: {e['msg']}"
                    )
                    for e in error.details["errors"]
                ]
            )

        elif error.code == "sql_execution_error" and "message" in error.details:
            level = "error"
            message = self.bad(error.details["message"])

        elif error.kind == "database" and error.code == "exception":
            level = "error"
            message = self.bad(error.details["message"])

        elif error.kind == "database" and error.code == "sayn_error":
            level = "error"
            message = self.bad(error.details["error_message"])

        elif error.kind == "database" and error.code == "introspection":
            level = "error"
            if "_message" in error.details["exception"].__dict__:
                message = self.bad(
                    f"Error during introspection: {error.details['exception']._message()}"
                )
            else:
                message = self.bad(
                    f"Error during introspection: {error.details['exception']}"
                )

        elif error.kind == "parsing" and "filename" in error.details:
            level = "error"
            if "error" in error.details:
                message = self.bad(
                    f"""Parsing error in file: {error.details['filename']}
                Details: {error.details['error']}
                   Line: {error.details['line']}"""
                )
            else:
                message = self.bad(f"File not found: {error.details['filename']}")

        elif error.kind == "task_type" and error.code == "invalid_task_type_error":
            level = "error"
            group = error.details["group"]
            task_type = error.details["type"]
            message = self.bad(
                f"""Task error in task group: {group}. Invalid task type: {task_type}.

            Current Valid Task Types:

            - autosql
            - sql
            - python
            - copy
            - dummy

            For more details please check SAYN documentation: https://173tech.github.io/sayn/tasks/overview/

            """
            )

        elif (
            error.kind == "python_loader"
            and error.code == "load_class_exception"
            and "exception" in error.details
        ):
            level = "error"
            message = self.bad(str(error.details["exception"]))

        elif (
            error.kind == "python_loader"
            and error.code == "missing_class"
            and "pyclass" in error.details
        ):
            level = "error"
            path = error.details["module_path"]
            if len(path) > 0:
                message = self.bad(
                    f"Error in file: {error.details['module_path']}.py. Missing Class: {error.details['pyclass']}."
                )
            else:
                message = self.bad(f"Invalid path: {error.details['pyclass']}")

        return {
            "level": level,
            "message": message,
        }

    # App context

    def app_start(self, details):
        debug = "(debug)" if details["debug"] else ""
        yesterday = date.today() - timedelta(days=1)
        if details["full_load"]:
            dt_range = "Full Load"
        elif (
            details["start_dt"] == details["end_dt"] and details["end_dt"] == yesterday
        ):
            dt_range = "Default"
        elif details["start_dt"] == details["end_dt"]:
            dt_range = f"{details['start_dt']}"
        else:
            dt_range = f"{details['start_dt']} to {details['end_dt']}"

        out = list()
        out.append(f"Starting sayn {debug}")
        out.append(f"Run ID: {details['run_id']}")
        out.append(f"Project: {details['project_name']}")
        out.append(f"Sayn version: {details['sayn_version']}")
        if details["project_git_commit"] is not None:
            out.append(f"Git commit: {details['project_git_commit']}")
        out.append(f"Period: {dt_range}")
        out.append(f"{'Profile: ' + (details.get('profile') or 'Default')}")

        return {"level": "info", "message": out}

    def app_finish(self, details):
        if "error" in details:
            return self.error_result(details["duration"], details["error"].error)
        else:
            errors = [
                t for t in list(details["tasks"].values()) if "FAILED" in str(t)
            ] + [t for t in list(details["tasks"].values()) if "SKIPPED" in str(t)]

            msg = f"Execution of SAYN took {human(details['duration'])}"
            if len(errors) > 0:
                return {"level": "error", "message": self.bad(msg)}
            else:
                return {"level": "info", "message": self.good(msg)}

    def app_stage_start(self, stage, details):
        if stage == "config":
            return {"level": "info", "message": "Configuring Project..."}
        elif stage == "setup":
            return {"level": "info", "message": "Setting up..."}
        elif stage in ("run", "compile", "test"):
            return {
                "level": "info",
                "message": self.bright(
                    f"Starting {stage} at {details['ts'].strftime('%H:%M')}..."
                ),
            }
        else:
            return self.unhandled("start_stage", "app", stage, details)

    def app_stage_finish(self, stage, details):
        tasks = group_list([(v.value, t) for t, v in details["tasks"].items()])
        failed = tasks.get("setup_failed", list()) + tasks.get("failed", list())
        succeeded = (
            tasks.get("ready", list())
            + tasks.get("ready_for_setup", list())
            + tasks.get("succeeded", list())
        )
        skipped = tasks.get("skipped", list())
        duration = human(details["duration"])
        totals_msg = (
            f"Total {'tasks' if details.get('test', False) is False else 'tests'}: {len(succeeded+failed+skipped)}. "
            f"Success: {len(succeeded)}. Failed {len(failed)}. Skipped {len(skipped)}."
        )
        if stage == "config":
            out = ["Finished project config:"]
            level = "info"
            if len(failed) > 0:
                out.append(self.bad(f"Tasks failed: {self.blist(failed)}"))
                level = "error"
            # if len(skipped) > 0:
            #     out.append(self.warn(f"Tasks to skip: {self.blist(skipped)}"))
            #     level = "error"
            if len(succeeded) > 0:
                out.append(
                    self.good(
                        f"{'Tasks' if details.get('test', False) is False else 'Tests'} found: {self.blist(succeeded)}"
                    )
                )
            return {"level": level, "message": out}

        elif stage == "setup":
            out = ["Finished setup:"]
            level = "info"
            type = "Tasks" if details.get("test", False) is False else "Tests"
            if len(failed) > 0:
                out.append(self.bad(f"{type} failed: {self.blist(failed)}"))
                level = "error"
            if len(skipped) > 0:
                out.append(self.warn(f"{type} to skip: {self.blist(skipped)}"))
                level = "error"
            if len(succeeded) > 0:
                out.append(self.good(f"{type} to run: {self.blist(succeeded)}"))
            return {"level": level, "message": out}

        elif stage in ("run", "compile", "test"):
            if len(failed) > 0 or len(skipped) > 0:
                out = [
                    self.red(
                        f"There were some errors during {stage} (took {duration})"
                    ),
                    self.red(totals_msg),
                ]
                if len(failed) > 0:
                    out.append(self.bad(f"Failed: {self.blist(failed)}"))
                if len(skipped) > 0:
                    out.append(self.warn(f"Skipped: {self.blist(skipped)}"))
                return {"level": "error", "message": out}
            else:
                return {
                    "level": "info",
                    "message": [
                        self.good(totals_msg),
                        self.good(
                            f"{stage.capitalize()} finished successfully in {duration}"
                        ),
                        f"Tasks executed: {self.blist(succeeded)}",
                    ],
                }

        else:
            return self.unhandled("finish_stage", "app", stage, details)

    # Task context

    def task_set_steps(self, details):
        return {
            "level": "info",
            "message": f"Run Steps: {self.blist(details['steps'])}",
        }

    def task_stage_start(self, stage, task, task_order, total_tasks, details):
        task_progress = f"[{task_order}/{total_tasks}]"
        ts = human(details["ts"])

        if stage == "config":
            return {"level": "info", "message": f"{self.bright(task)}"}
        elif stage == "setup":
            return {"level": "info", "message": f"{task_progress} {self.bright(task)}"}
        elif stage in ("run", "compile", "test"):
            return {
                "level": "info",
                "message": f"{self.bright(task_progress +' ' +task)} (started at {ts})",
            }
        else:
            return self.unhandled("start_stage", "task", stage, details)

    def task_stage_finish(self, stage, task, task_order, total_tasks, details):
        duration = human(details["duration"])

        if details.get("result") is None or details["result"].is_ok:
            return {
                "level": "info",
                "message": self.good(f"Took ({duration})"),
            }
        else:
            return self.error_result(details["duration"], details["result"].error)

    def task_step_start(self, stage, task, step, step_order, total_steps, details):
        task_progress = f"[{step_order}/{total_steps}]"
        ts = f"[{human(details['ts'])}]" if self.output_ts else ""

        if stage in ("run", "compile", "test"):
            return {
                "level": "info",
                "message": self.info(
                    f"{task_progress} {ts} Executing {self.bright(step)}"
                ),
            }
        else:
            return self.unhandled("start_step", "task", stage, details)

    def task_step_finish(self, stage, task, step, step_order, total_steps, details):
        task_progress = f"[{step_order}/{total_steps}]"
        ts = f"[{human(details['ts'])}]" if self.output_ts else ""
        duration = human(details["duration"])

        if details["result"].is_ok:
            return {
                "level": "info",
                "message": self.good(
                    f"{task_progress}" + f" {ts} {self.bright(step)} ({duration})"
                ),
            }
        else:
            return self.error_result(details["duration"], details["result"].error)



from sqlalchemy import create_engine, event
from itertools import groupby

from typing import Optional

from . import Database

db_parameters = ["database"]


class Sqlite(Database):
    def feature(self, feature):
        return feature in (
            "CANNOT ALTER INDEXES",
            "CANNOT SET SCHEMA",
            "CANNOT SPECIFY DDL IN SELECT",
            "NO SCHEMA SUPPORT",
        )

    def create_engine(self, settings):
        database = settings.pop("database")

        engine = create_engine(f"sqlite:///{database}", **settings)

        # this is set to fix a SQLite setting which can prevent a second execution of SAYN.
        # More info on this command here: https://sqlite.org/pragma.html#pragma_legacy_alter_table
        @event.listens_for(engine, "connect")
        def do_connect(dbapi_connection, connection_record):
            cursor = dbapi_connection.cursor()
            cursor.execute("PRAGMA legacy_alter_table = ON")
            cursor.close()

        return engine

    def _list_databases(self):
        return [""]

    def _get_table_type(self, type):
        if type == "table":
            return "table"
        elif type == "view":
            return "view"

    def _list_objects(self, databases):
        """List the accessible databases for this connection."""
        objects = list()
        queries = ""
        for database in databases:
            if database == "":

                queries += """SELECT "" AS table_schema
                                    , "" AS table_catalog
                                    , name AS table_name
                                    , type AS table_type
                                 FROM (SELECT * FROM sqlite_schema UNION ALL
                                 SELECT * FROM sqlite_temp_schema)
                            """
            queries += "\nUNION ALL\n"

        queries = "\n".join(queries.split("\n")[:-2])
        objects = {
            db.lower(): {
                schema.lower(): {
                    table.lower(): {
                        "type": self._get_table_type(type["table_type"]) for type in ggg
                    }
                    for table, ggg in groupby(gg, lambda x: x["table_name"])
                }
                for schema, gg in groupby(g, lambda x: x["table_schema"])
            }
            for db, g in groupby(
                sorted(
                    self.read_data(queries),
                    key=lambda x: (
                        x["table_catalog"],
                        x["table_schema"],
                        x["table_name"],
                    ),
                ),
                key=lambda x: x["table_catalog"],
            )
        }

        return objects

    def execute(self, script):
        with self.engine.connect().execution_options(autocommit=True) as connection:
            connection.connection.executescript(script)

    def _obj_str(
        self, database: Optional[str], schema: Optional[str], table: str
    ) -> str:
        if schema is not None:
            raise ValueError("Sqlite doesn't support schemas")

        if database is not None:
            raise ValueError("Sqlite doesn't support databases")

        return (
            f"{database + '.' if database is not None else ''}"
            f"{schema + '.' if schema is not None else ''}"
            f"{table}"
        )

    def _fully_qualify(
        self, database: Optional[str], schema: Optional[str], table: str
    ) -> str:
        return self._obj_str(database, schema, table)



from pathlib import Path
from typing import Any, List, Mapping, Optional, Union
from enum import Enum
import re

from pydantic import BaseModel, FilePath, validator, Extra

from ..core.errors import Exc, Ok, Err
from ..database import Database
from .task import Task


class BaseConfig(BaseModel):
    supports_schemas: bool
    db_type: str

    delete_key: Optional[str]
    destination: Optional[str]
    materialisation: Optional[str]
    db: Optional[str]
    tmp_db: Optional[str]
    tmp_schema: Optional[str]

    columns: Optional[List[Union[str, Mapping[str, Any]]]] = list()
    table_properties: Mapping[str, Any] = dict()
    post_hook: List[Mapping[str, Any]] = list()

    class Config:
        extra = Extra.forbid


class Config(BaseConfig):
    sql_folder: Path
    file_name: FilePath

    class Config:
        extra = Extra.forbid

    @validator("file_name", pre=True)
    def file_name_plus_folder(cls, v, values):
        return Path(values["sql_folder"], v)

    @validator("tmp_schema")
    def can_use_tmp_schema(cls, v, values):
        if v is not None and not values["supports_schemas"]:
            raise ValueError(
                f'tmp_schema not supported for database of type {values["db_type"]}'
            )

        return v

    @validator("materialisation")
    def incremental_has_delete_key(cls, v, values):
        if v not in ("table", "view", "incremental", "script", ""):
            raise ValueError(
                f'"{v}". Valid materialisations: table, view, incremental, script'
            )
        elif v != "incremental" and values.get("delete_key") is not None:
            raise ValueError('"delete_key" is invalid in non-incremental loads')
        elif v == "incremental" and values.get("delete_key") is None:
            raise ValueError('"delete_key" is required for incremental loads')
        elif v != "script" and values.get("destination") is None:
            raise ValueError('"destination" field is required')
        elif v == "script" and values.get("destination") is not None:
            raise ValueError('"destination" is invalid for script materialisation')
        elif v == "":
            return "script"
        else:
            return v

    @validator("destination")
    def can_use_schema(cls, v, values):
        match = re.match(r"(.*)\.(.*)|(\w+)", v)
        if match.group(1) is not None and not values["supports_schemas"]:
            raise ValueError(
                f'schema not supported for database of type {values["db_type"]}'
            )

        return v


class OnFailValue(str, Enum):
    skip = "skip"
    no_skip = "no_skip"


class CompileConfig(BaseConfig):
    tags: Optional[List[str]]
    sources: Optional[List[str]]
    outputs: Optional[List[str]]
    parents: Optional[List[str]]
    on_fail: Optional[OnFailValue]

    class Config:
        extra = Extra.forbid


class SqlTask(Task):
    @property
    def target_db(self):
        return self.connections[self._target_db]

    def config(self, **config):

        def_connections_src = []
        def_connections_out = []

        def def_src(obj, level=None):
            def_connections_src.append((obj, level))
            return ""

        def def_out(obj, level=None):
            def_connections_out.append((obj, level))
            return ""

        if "task_name" in self._config_input:
            del self._config_input["task_name"]

        # if "columns" in config:
        #     self._has_tests = True

        conn_names_list = [
            n for n, c in self.connections.items() if isinstance(c, Database)
        ]

        # set the target db for execution
        # this check needs to happen here so we can pass db_features and db_type to the validator
        if isinstance(config.get("db"), str):
            if config["db"] not in conn_names_list:
                return Err(
                    "task_definition",
                    "destination_db_not_in_settings",
                    db=config["db"],
                )
            self._target_db = config["db"]
        else:
            self._target_db = self._default_db

        config.update(
            {
                "supports_schemas": not self.target_db.feature("NO SCHEMA SUPPORT"),
                "db_type": self.target_db.db_type,
            }
        )

        try:
            self.task_config = Config(
                sql_folder=self.run_arguments["folders"]["sql"], **config
            )
        except Exception as e:
            return Exc(e)

        # We compile first to allow changes to the config
        # Template compilation
        self.compiler.update_globals(
            src=lambda x, level=None: def_src(x, level=level),
            out=lambda x, level=None: def_out(x, level=level),
            config=self.config_macro,
        )
        self.allow_config = True

        self.prepared_sql_query = self.compiler.prepare(self.task_config.file_name)
        self.sql_query = self.prepared_sql_query.compile()

        # After the first compilation, config shouldn't be executed again
        self.allow_config = False

        if isinstance(config.get("destination"), str):
            match = re.match(
                r"(.*)\.(.*)\.(.*)|(.*)\.(.*)|(\w+)", self.task_config.destination
            )

            if match.group(1):
                database_name = match.group(1)
                db_schema = match.group(2)
                base_table_name = match.group(3)
            elif match.group(4):
                database_name = None
                db_schema = match.group(4)
                base_table_name = match.group(5)
            else:
                database_name = None
                db_schema = None
                base_table_name = match.group(6)
            # Output calculation
            if self.task_config.tmp_schema is not None:
                tmp_db_schema = self.task_config.tmp_schema
            else:
                tmp_db_schema = db_schema

            if self.task_config.tmp_db is not None:
                tmp_db = self.task_config.tmp_db
            else:
                tmp_db = database_name

            if (database_name is None) and (db_schema is None):
                self.database = None
                self.schema = None
                self.table = self.out(base_table_name, self.target_db)
            elif (database_name is None) and (db_schema is not None):
                obj = self.out(f"{db_schema}.{base_table_name}", self.target_db)
                self.database = None
                self.schema = obj.split(".")[0]
                self.table = obj.split(".")[1]
            else:
                obj = self.out(
                    f"{database_name}.{db_schema}.{base_table_name}", self.target_db
                )
                self.database = obj.split(".")[0]
                self.schema = obj.split(".")[1]
                self.table = obj.split(".")[2]

            if (tmp_db is None) and (tmp_db_schema is None):
                self.tmp_db = None
                self.tmp_schema = None
                self.tmp_table = self.out(f"sayn_tmp_{base_table_name}", self.target_db)
            elif (tmp_db is None) and (tmp_db_schema is not None):
                obj = self.out(
                    f"{tmp_db_schema}.sayn_tmp_{base_table_name}", self.target_db
                )
                self.tmp_db = None
                self.tmp_schema = obj.split(".")[0]
                self.tmp_table = obj.split(".")[1]
            else:
                obj = self.out(
                    f"{tmp_db}.{tmp_db_schema}.sayn_tmp_{base_table_name}",
                    self.target_db,
                )
                self.tmp_db = obj.split(".")[0]
                self.tmp_schema = obj.split(".")[1]
                self.tmp_table = obj.split(".")[2]

        self.materialisation = self.task_config.materialisation
        self.delete_key = self.task_config.delete_key

        self.compiler.update_globals(
            src=lambda x, level=None: self.src(
                x, connection=self._target_db, level=level
            ),
            out=lambda x, level=None: self.out(
                x, connection=self._target_db, level=level
            ),
            config=self.config_macro,
        )

        for s in def_connections_src:
            self.src(s[0], connection=self._target_db, level=s[1])

        for o in def_connections_out:
            self.out(o[0], connection=self._target_db, level=o[1])

        # DDL validation
        result = self.target_db._validate_ddl(
            self.task_config.columns,
            self.task_config.table_properties,
            self.task_config.post_hook,
        )
        if result.is_err:
            return result
        else:
            self.ddl = result.value

        if self.run_arguments["command"] == "test" and len(self.ddl["columns"]) != 0:
            result = self.target_db._construct_tests(
                self.ddl["columns"], self.table, self.schema
            )
            if result.is_err:
                return result
            else:
                self.test_query = result.value[0]
                self.test_breakdown = result.value[1]

            if self.test_query is not None:
                self._has_tests = True

        return Ok()

    def config_macro(self, **config):
        if self.allow_config:
            config.update(
                {
                    "supports_schemas": not self.target_db.feature("NO SCHEMA SUPPORT"),
                    "db_type": self.target_db.db_type,
                }
            )
            task_config_override = CompileConfig(**config)

            self.task_config.materialisation = (
                task_config_override.materialisation or self.task_config.materialisation
            )

            self.task_config.db = (
                task_config_override.db or self.task_config.db or self._default_db
            )

            conn_names_list = [
                n for n, c in self.connections.items() if isinstance(c, Database)
            ]

            if self.task_config.db not in conn_names_list:
                return Err(
                    "task_definition",
                    "destination_db_not_in_settings",
                    db=config["db"],
                )
            else:
                self._target_db = self.task_config.db

            self.task_config.tmp_schema = (
                task_config_override.tmp_schema or self.task_config.tmp_schema
            )

            self.task_config.destination = (
                task_config_override.destination or self.task_config.destination
            )

            self.task_config.columns = (
                task_config_override.columns or self.task_config.columns
            )
            self.task_config.table_properties = (
                task_config_override.table_properties
                or self.task_config.table_properties
            )
            self.task_config.post_hook = (
                task_config_override.post_hook or self.task_config.post_hook
            )
            self.task_config.delete_key = (
                task_config_override.delete_key or self.task_config.delete_key
            )
            # Sent to the wrapper
            if task_config_override.on_fail is not None:
                self._config_input["on_fail"] = task_config_override.on_fail

            if task_config_override.tags is not None:
                self._config_input["tags"] = task_config_override.tags

            if task_config_override.sources is not None:
                self._config_input["sources"] = task_config_override.sources

            if task_config_override.outputs is not None:
                self._config_input["outputs"] = task_config_override.outputs

            if task_config_override.parents is not None:
                self._config_input["parents"] = task_config_override.parents
        # Returns an empty string to avoid productin incorrect sql
        return ""

    def setup(self):
        # recompile regardless
        self.sql_query = self.prepared_sql_query.compile()

        if self._has_tests and self._needs_recompile:
            obj = self.src(f"{self.task_config.destination}", self.target_db)
            test_schema = obj.split(".")[0]
            test_table = obj.split(".")[1]
            result = self.target_db._construct_tests(
                self.ddl["columns"], test_table, test_schema
            )
            if result.is_err:
                return result
            else:
                self.test_query = result.value[0]
                self.test_breakdown = result.value[1]

        return Ok()

    def execute(self, execute, debug):
        if self.materialisation in ("table", "view", "incremental"):
            if self.run_arguments["debug"]:
                self.write_compilation_output(self.sql_query, "select")
            else:
                self.write_compilation_output(self.sql_query)
        elif self.materialisation in ("script",):
            self.write_compilation_output(self.sql_query)

        if self.materialisation == "view":
            # View
            step_queries = self.target_db.replace_view(
                self.table,
                self.sql_query,
                schema=self.schema,
                db=self.database,
                **self.ddl,
            )

        elif self.materialisation == "table":
            # Full load or target table missing
            if self.target_db.feature("CANNOT CHANGE SCHEMA"):
                # Use destination schema if the db doesn't support schema changes
                tmp_schema = self.schema
                tmp_db = self.database
            else:
                tmp_schema = self.tmp_schema
                tmp_db = self.tmp_db

            step_queries = self.target_db.replace_table(
                self.table,
                self.sql_query,
                schema=self.schema,
                db=self.database,
                tmp_schema=tmp_schema,
                tmp_db=tmp_db,
                **self.ddl,
            )

        elif self.materialisation == "incremental":
            # Incremental load
            step_queries = self.target_db.merge_query(
                self.table,
                self.sql_query,
                self.delete_key,
                schema=self.schema,
                db=self.database,
                tmp_schema=self.tmp_schema,
                **self.ddl,
            )
        else:
            # script
            step_queries = {"Write Query": None, "Execute Query": self.sql_query}

        self.set_run_steps(list(step_queries.keys()))

        for step, query in step_queries.items():
            with self.step(step):
                if debug and query:
                    self.write_compilation_output(query, step.replace(" ", "_").lower())
                if execute and query:
                    try:
                        self.target_db.execute(query)
                    except Exception as e:
                        return Exc(e)

        return Ok()

    def compile(self):
        return self.execute(False, self.run_arguments["debug"])

    def run(self):
        return self.execute(True, self.run_arguments["debug"])

    def test(self):
        step_queries = {
            "Write Test Query": self.test_query,
            "Execute Test Query": self.test_query,
        }
        breakdown = self.get_test_breakdown(self.test_breakdown)

        if self.test_query is None:
            self.info("Nothing to be done")
            return self.success()
        else:
            self.set_run_steps(list(step_queries.keys()))

            for step, query in step_queries.items():
                with self.step(step):
                    if "Write" in step:
                        self.write_compilation_output(query, "test")
                    if "Execute" in step:
                        try:
                            result = self.default_db.read_data(query)
                        except Exception as e:
                            return Exc(e)

            if len(result) == 0:
                return self.test_sucessful(breakdown)
            else:
                errout, failed = self.test_failure(
                    breakdown, result, self.run_arguments["debug"]
                )
                problematic_values_query = self.default_db.test_problematic_values(
                    failed, self.table, self.schema
                )

                for query in problematic_values_query.split(";"):
                    if query.strip():
                        header = re.search(r"--.*?--", query).group(0)
                        self.info("")
                        self.info(header)
                        self.info(
                            "===================================================================="
                        )
                        self.info(
                            re.sub(r"--.*?--", "", query).replace("\n", " ").strip()
                            + ";"
                        )
                        self.info(
                            "===================================================================="
                        )
                        self.info("")

                self.write_compilation_output(
                    problematic_values_query, "test_problematic_values"
                )

                return errout



from copy import deepcopy
from collections import Counter
import csv
import datetime
import decimal
from itertools import groupby
import io
from typing import List, Optional, Union
from uuid import UUID

import orjson
from pydantic import validator, Extra, BaseModel
from sqlalchemy import create_engine
from sqlalchemy.sql import sqltypes

from . import Database, Columns, Hook, BaseDDL

from ..core.errors import DBError, Ok

db_parameters = ["project", "credentials_path", "location", "dataset"]


class DDL(BaseDDL):
    class Properties(BaseModel):
        partition: Optional[str]
        cluster: Optional[List[str]]

        class Config:
            extra = Extra.forbid

    columns: List[Union[str, Columns]] = list()
    properties: Optional[Properties]
    post_hook: List[Hook] = list()

    class Config:
        extra = Extra.forbid

    @validator("columns", pre=True)
    def transform_str_cols(cls, v):
        if v is not None and isinstance(v, List):
            return [{"name": c} if isinstance(c, str) else c for c in v]
        else:
            return v

    @validator("columns")
    def columns_unique(cls, v):
        dupes = {k for k, v in Counter([e.name for e in v]).items() if v > 1}
        if len(dupes) > 0:
            raise ValueError(f"Duplicate columns: {','.join(dupes)}")
        else:
            return v

    @validator("properties")
    def validate_properties(cls, v, values):
        if v is not None and v.cluster is not None:
            if len(values.get("columns")) > 0:
                missing_columns = set(v) - set([c.name for c in values.get("columns")])
                if len(missing_columns) > 0:
                    raise ValueError(
                        f'Cluster contains columns not specified in the ddl: "{missing_columns}"'
                    )

        return v

    def get_ddl(self):
        result = self.base_ddl()
        properties = list()
        if self.properties is not None:
            if self.properties.cluster is not None:
                properties.append({"cluster": self.properties.cluster})
                result["cluster"] = self.properties.cluster

            if self.properties.partition is not None:
                properties.append({"partition": self.properties.partition})
                result["partition"] = self.properties.partition

        return result


class Bigquery(Database):
    DDL = DDL

    project = None
    dataset = None
    client = None

    def feature(self, feature):
        return feature in (
            "CAN REPLACE TABLE",
            "CAN REPLACE VIEW",
            "CANNOT CHANGE SCHEMA",
        )

    def create_engine(self, settings):
        settings = deepcopy(settings)
        self.project = settings.pop("project")

        url = f"bigquery://{self.project}"
        if "dataset" in settings:
            self.dataset = settings.pop("dataset")
            url += "/" + self.dataset

        engine = create_engine(url, **settings)
        self.client = engine.raw_connection()._client

        return engine

    def _construct_tests(self, columns, table, schema=None):
        count_tests, query, breakdown = self._construct_tests_template(
            columns, table, "standard_tests_bigquery.sql", schema
        )
        if count_tests == 0:
            return Ok([None, breakdown])

        return Ok([query, breakdown])

    def test_problematic_values(self, failed: list, table: str, schema: str) -> str:
        return self.test_problematic_values_template(
            failed, table, schema, "standard_test_output_bigquery.sql"
        )

    def _introspect(self, objects_to_introspect):
        objects = list()

        def get_project_key(x):
            return "" if x == self.project else x

        def get_dataset_key(x):
            return "" if x == self.dataset else x

        # List all project we have access to that are
        # in use in the project
        objects = {
            project.project_id: {
                "is_default_project": project.project_id == self.project,
                "datasets": dict(),
            }
            for project in self.client.list_projects()
            if project.project_id in objects_to_introspect
            or project.project_id == self.project
        }

        for project_id in objects.keys():
            project = objects[project_id]

            if get_project_key(project_id) not in objects_to_introspect:
                continue

            requested_datasets = list(
                objects_to_introspect[get_project_key(project_id)].keys()
            )
            objects[project_id]["datasets"] = {
                dataset.dataset_id: {
                    "is_default_dataset": project["is_default_project"]
                    and dataset.dataset_id == self.dataset,
                    "objects": dict(),
                }
                for dataset in self.client.list_datasets(project=project_id)
                if dataset.dataset_id in requested_datasets
                or (
                    project["is_default_project"] and dataset.dataset_id == self.dataset
                )
            }

            for dataset_id in project["datasets"].keys():
                dataset = project["datasets"][dataset_id]
                query = f"""SELECT t.table_name
                                 , t.table_type
                                 , array_agg(STRUCT(c.column_name, c.is_partitioning_column = 'YES' AS is_partition, c.clustering_ordinal_position)
                                              ORDER BY clustering_ordinal_position) AS columns
                              FROM `{project_id}`.{dataset_id}.INFORMATION_SCHEMA.TABLES t
                              JOIN `{project_id}`.{dataset_id}.INFORMATION_SCHEMA.COLUMNS c
                                ON c.table_name = t.table_name
                             GROUP BY 1,2
                            """

                result = self.read_data(query)

                dataset["objects"] = {
                    table["table_name"]: self._get_table_type(table) for table in result
                }

        self._requested_objects = {
            get_project_key(project): {
                get_dataset_key(dataset): dataset_details["objects"]
                for dataset, dataset_details in project_details["datasets"].items()
            }
            for project, project_details in objects.items()
        }

    def _get_table_type(self, type):
        out_key = list()
        out_value = list()

        if type["table_type"] == "BASE TABLE":
            out_key.append("type")
            out_value.append("table")
        elif type["table_type"] == "VIEW":
            out_key.append("type")
            out_value.append("view")

        cluster_cols = []
        for c in type["columns"]:
            if c["is_partition"] is True:
                out_key.append("partition")
                out_value.append(c["column_name"])
            if c["clustering_ordinal_position"] is not None:
                cluster_cols.append(c["column_name"])
        if cluster_cols:
            out_key.append("cluster")
            out_value.append(cluster_cols)

        return dict(zip(out_key, out_value))

    def _py2sqa(self, from_type):
        python_types = {
            int: sqltypes.Integer,
            str: sqltypes.String,
            float: sqltypes.Float,
            decimal.Decimal: sqltypes.Numeric,
            datetime.datetime: sqltypes.TIMESTAMP,
            bytes: sqltypes.LargeBinary,
            bool: sqltypes.Boolean,
            datetime.date: sqltypes.Date,
            datetime.time: sqltypes.Time,
            datetime.timedelta: sqltypes.Interval,
            list: sqltypes.ARRAY,
            dict: sqltypes.JSON,
            UUID: sqltypes.String,
        }

        if from_type not in python_types:
            raise ValueError(f'Type not supported "{from_type}"')
        else:
            return python_types[from_type]().compile(dialect=self.engine.dialect)

    def _load_data_batch(self, table, data, schema, db):
        full_table_name = f"{self.project if db is None else db}.{self.dataset if schema is None else schema}.{table}"

        from google.cloud import bigquery

        job_config = bigquery.LoadJobConfig(
            source_format=bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
        )
        client = self.engine.raw_connection()._client

        data_bytes = b"\n".join([orjson.dumps(d) for d in data])
        job = client.load_table_from_file(
            io.BytesIO(data_bytes), full_table_name, job_config=job_config
        )
        job.result()

    def move_table(
        self,
        src_table,
        dst_table,
        src_schema=None,
        dst_schema=None,
        src_db=None,
        dst_db=None,
        **ddl,
    ):
        # ddl = self._format_properties(ddl).value

        full_src_table = f"{src_db + '.' if src_schema is not None else ''}{src_schema + '.' if src_schema is not None else ''}{src_table}"
        select = f"SELECT * FROM {full_src_table}"
        create_or_replace = self.create_table(
            dst_table, dst_schema, dst_db, select=select, replace=True, **ddl
        )

        return "\n\n".join((create_or_replace, f"DROP TABLE {full_src_table}"))

    def create_table(
        self,
        table,
        schema=None,
        db=None,
        select=None,
        replace=False,
        **ddl,
    ):
        db_name = db or ""
        schema_name = schema or ""
        full_name = fully_qualify(table, schema, db)
        if (
            db_name in self._requested_objects
            and schema_name in self._requested_objects[db_name]
            and table in self._requested_objects[db_name][schema_name]
        ):
            db_info = self._requested_objects[db_name][schema_name][table]
            object_type = db_info.get("type")
            table_exists = bool(object_type == "table")
            view_exists = bool(object_type == "view")
            partition_column = db_info.get("partition", "")
            cluster_column = set(db_info.get("cluster", set()))
        else:
            db_info = dict()
            table_exists = True
            view_exists = True
            partition_column = ""
            cluster_column = set()

        des_partitioned = ddl.get("partition") or ""
        des_clustered = set(ddl.get("cluster") or set())

        if des_clustered == cluster_column and des_partitioned == partition_column:
            drop = ""
        elif db_info.get("type") == "table":
            drop = f"DROP TABLE IF EXISTS {full_name};\n"
        else:
            drop = f"DROP VIEW IF EXISTS {full_name};\n"

        template = self._jinja_env.get_template("create_table.sql")
        query = template.render(
            table_name=table,
            full_name=full_name,
            view_exists=view_exists,
            table_exists=table_exists,
            select=select,
            replace=True,
            can_replace_table=self.feature("CAN REPLACE TABLE"),
            needs_cascade=self.feature("NEEDS CASCADE"),
            cannot_specify_ddl_select=self.feature("CANNOT SPECIFY DDL IN SELECT"),
            all_columns_have_type=len(
                [c for c in ddl.get("columns", dict()) if c.get("type") is not None]
            ),
            **ddl,
        )
        query = drop + query
        return query


def fully_qualify(name, schema=None, db=None):
    return f"{db+'.' if db is not None else ''}{schema+'.' if schema is not None else ''}{name}"



import os
from pathlib import Path
import shutil


def sayn_init(sayn_project_name):
    src = Path(
        os.path.dirname(os.path.abspath(__file__)), "data", "init_project"
    ).absolute()
    dst = Path(sayn_project_name).absolute()

    print(
        "Creating SAYN base project at the following location: {dst}...".format(dst=dst)
    )
    try:
        shutil.copytree(src, dst)

        # rename the sample_settings.yaml to settings.yaml
        os.rename(
            Path(dst, "sample_settings.yaml").absolute(),
            Path(dst, "settings.yaml").absolute(),
        )

        print("SAYN base project created.")
        print("For more information about SAYN visit our documentation:")
        print("https://173tech.github.io/sayn/")
    except OSError as e:
        print(e)



from .bigquery import Bigquery
from .mysql import Mysql
from .postgresql import Postgresql
from .redshift import Redshift
from .sqlite import Sqlite
from .snowflake import Snowflake
from .unknown import UnknownDb

drivers = {
    "postgresql": Postgresql,
    "sqlite": Sqlite,
    "mysql": Mysql,
    "snowflake": Snowflake,
    "redshift": Redshift,
    "bigquery": Bigquery,
}

db_params = ("max_batch_rows", "type")


def create(name, name_in_settings, settings):
    db_type = settings.pop("type")
    if db_type not in drivers:
        raise ValueError(f"No driver for {db_type} found")
    else:
        # Extract common parameters
        common_params = {k: v for k, v in settings.items() if k in db_params}
        settings = {k: v for k, v in settings.items() if k not in db_params}

        db_obj = drivers[db_type](
            name,
            name_in_settings,
            db_type,
            common_params,
            settings,
        )

        return db_obj


def create_dummy(name):
    return UnknownDb(name, name, "dummy", dict(), dict())



from datetime import datetime

from ..core.errors import Ok


class TaskEventTracker:
    # _logger = None
    # _task_name = None
    # _task_order = None
    # _steps = None
    # _current_step = None
    # _current_step_start_ts = None

    def __init__(self, logger, task_name, task_order):
        self._logger = logger
        self._task_name = task_name
        self._task_order = task_order
        self._steps = list()
        self._current_step = None
        self._current_step_start_ts = None

    def _report_event(self, event, **details):
        details["event"] = event
        details["context"] = "task"

        details["task"] = self._task_name
        details["task_order"] = self._task_order

        details["step"] = self._current_step
        details["step_order"] = (
            self._steps.index(self._current_step) + 1
            if self._current_step in self._steps
            else None
        )
        details["total_steps"] = len(self._steps)

        self._logger.report_event(**details)

    def set_run_steps(self, steps):
        self._report_event("set_run_steps", steps=steps)
        self._steps = steps

    def add_run_steps(self, steps):
        if isinstance(steps, str):
            steps = [steps]
        self._report_event("set_run_steps", steps=steps)
        self._steps.extend(steps)

    def start_step(self, step):
        self.finish_current_step()

        if step not in self._steps:
            self._steps.append(step)

        self._current_step = step
        self._current_step_start_ts = datetime.now()

        self._report_event("start_step")

    def finish_current_step(self, result=Ok()):
        if self._current_step is not None:
            if self._current_step_start_ts is not None:
                duration = datetime.now() - self._current_step_start_ts
            else:
                duration = None

            self._report_event(
                "finish_step",
                duration=duration,
                result=result,
            )
            self._current_step = None
            self._current_step_start_ts = None

    def debug(self, message, **details):
        self._report_event("message", level="debug", message=message, **details)

    def info(self, message, **details):
        self._report_event("message", level="info", message=message, **details)

    def warning(self, message, **details):
        self._report_event("message", level="warning", message=message, **details)

    def error(self, message, **details):
        self._report_event("message", level="error", message=message, **details)



from typing import Any, Dict, Optional, Sequence

from pydantic import ValidationError
from ruamel.yaml.error import MarkedYAMLError
import sqlalchemy
from sqlalchemy.exc import SQLAlchemyError


class Error:
    kind: str
    code: str
    details: Dict

    def __init__(self, kind, code, details):
        self.kind = kind
        self.code = code
        self.details = details

    def __repr__(self):
        return f"Result.Err ({self.kind}::{self.code}): {self.details.__repr__()}"


class Result:
    is_ok: bool = False
    value: Any = None
    error: Optional[Error] = None

    def __init__(self, value: Any = None, error: Error = None):
        if error is not None:
            self.is_ok = False
            self.error = error
        else:
            self.is_ok = True
            self.value = value

    def __repr__(self):
        if self.is_ok:
            return f"Result.Ok: {self.value.__repr__()}"
        else:
            return self.error.__repr__()

    @property
    def is_err(self):
        return not self.is_ok


def Ok(value=None):
    """Creates an OK result"""
    return Result(value=value)


def Err(kind, code, **details):
    """Creates an Error result"""
    return Result(error=Error(kind, code, details))


def Exc(exc, **kwargs):
    """Creates an Error result from an exception"""
    if isinstance(exc, MarkedYAMLError):
        # Ruamel error
        # TODO check that it's controlled from sayn.core.config
        return Result(
            error=Error(
                "parsing",
                "yaml_parse",
                dict(
                    **kwargs,
                    **{"error": exc.problem, "line": exc.problem_mark.line + 1},
                ),
            )
        )

    elif isinstance(exc, ValidationError):
        # Pydantic error
        return Result(
            error=Error("parsing", "validation_error", {"errors": exc.errors()})
        )

    elif isinstance(exc, NotImplementedError) and exc.args[0] == "SAYN task":
        # Missing implementation for SAYN command
        return Result(
            error=Error(
                "exception",
                "not_implemented",
                {"class": exc.args[1], "method": exc.args[2], "exception": exc},
            )
        )
    elif isinstance(exc, NotImplementedError):
        # Other missing method
        return Result(
            error=Error("exception", "unknown_not_implemented", {"exception": exc})
        )

    elif isinstance(exc, SaynError):
        return Err(**exc.payload())

    # TODO add other exceptions like:
    #   - DB errors
    #   - Jinja
    elif isinstance(exc, SQLAlchemyError):
        return Result(
            error=Error(
                "database",
                "exception",
                {"exception": exc, "message": " ".join(exc.args)},
            )
        )
    else:
        return Result(
            error=Error("exception", "unhandled_exception", {"exception": exc})
        )


class SaynError(Exception):
    def payload(self):
        return dict()


class SaynCompileError(SaynError):
    def __init__(self, value):
        self.value = value

    def payload(self):
        return {"value": self.value}


class SaynMissingFileError(SaynError):
    def __init__(self, filename, is_folder=False):
        self.file_name = filename
        self.is_folder = is_folder

    def payload(self):
        return {
            "kind": "missing_file",
            "code": "missing_folder",
            "error_message": f"Missing {'folder' if self.is_folder else 'file'} \"{self.file_name}\"",
            "file_name": self.file_name,
        }


class SaynParsingError(SaynError):
    errors: Sequence

    def __init__(self, code, errors):
        self.code = code
        self.errors = errors

    def payload(self):
        # Sort by file_name to compress the message output
        sorted_errors = sorted(self.errors, key=lambda x: x["file_name"])

        file_name = sorted_errors[0]["file_name"]
        message = f'Error in file "{file_name}"'
        for error in sorted_errors:
            if error["file_name"] != file_name:
                # Add a new file_name line if it's different
                message += f'\nIn file "{file_name}"'

            if "snippet" in error:
                message += "\n" + error["snippet"]
            elif "loc" in error:
                message += (
                    f"\n  In \"{' > '.join([str(item) for item in error['loc']])}\""
                    f" (line {error['line']}): {error['message']}"
                )
            else:
                message += "\n" + error["message"]

        return {
            "kind": "parsing_error",
            "code": self.code,
            "error_message": message,
            "errors": self.errors,
        }


class DagCycleError(SaynError):
    def __init__(self, cycle):
        self.cycle = cycle


class DBError(SaynError):
    db_name = None
    db_type = None

    def __init__(self, db_name, db_type, message, errors=None):
        self.db_name = db_name
        self.db_type = db_type
        self.message = message
        self.errors = errors

    def payload(self):
        return {
            "kind": "database",
            "code": "sayn_error",
            "error_message": f"Error on connection {self.db_name}: {self.message}",
        }



from copy import deepcopy
import os
from pathlib import Path
import re
from typing import Any, Mapping, Optional, Sequence

from pydantic import BaseModel, validator, ValidationError, Extra
from ruamel.yaml import YAML
from ruamel.yaml.error import MarkedYAMLError

from ..database.creator import create as create_db, create_dummy
from .errors import Err, Exc, Ok

RE_ENV_VAR_NAME = re.compile(
    r"SAYN_((?P<stringify>(SCHEMA|TABLE)_(PREFIX|SUFFIX|STRINGIFY))"
    r"|(?P<from_prod>FROM_PROD)$"
    r"|(?P<default_run>DEFAULT_RUN)$"
    r"|(?P<type>PARAMETER|CREDENTIAL)_(?P<name>.+))$"
)

RE_DEFAULT_RUN_VAL = re.compile(
    r"^( *(?:(?:-t|--tasks|-x|--exclude)(?: +(?:group\:|tag\:)?[a-zA-Z][a-zA-Z_0-9]+)+|-u|--upstream-prod))+$"
)
RE_DEFAULT_RUN = re.compile(
    r" *((?:-t|--tasks|-x|--exclude)(?: +(?:group\:|tag\:)?[a-zA-Z][a-zA-Z_0-9]+)+|-u|--upstream-prod)"
)


class TableGlob(str):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, str):
            raise TypeError("string required")
        m = re.match(
            r"((?P<schema>[a-zA-Z\*][_0-9a-zA-Z\*]*).)?(?P<table>[a-zA-Z\*][_0-9a-zA-Z\*]*)",
            v,
        )
        if not m:
            raise ValueError("invalid table specification")

        out = ""
        if m.groupdict()["schema"] is not None:
            out += m.groupdict()["schema"].replace("*", ".*") + r"\."
        if m.groupdict()["table"] is not None:
            out += m.groupdict()["table"].replace("*", ".*")

        return cls(out)


class Environment(BaseModel):
    class Stringify(BaseModel):
        database_prefix: Optional[str]
        database_suffix: Optional[str]
        database_stringify: Optional[str]
        schema_prefix: Optional[str]
        schema_suffix: Optional[str]
        schema_override: Optional[str]
        table_prefix: Optional[str]
        table_suffix: Optional[str]
        table_override: Optional[str]

    parameters: Optional[Mapping[str, Any]]
    credentials: Optional[Mapping[str, Mapping[str, Any]]]
    stringify: Optional[Stringify]
    from_prod: Optional[Sequence[TableGlob]]
    default_run: Optional[str]

    class Config:
        extra = Extra.forbid
        anystr_lower = True

    @validator("default_run")
    def default_run_validator(cls, v):
        m = RE_DEFAULT_RUN_VAL.match(v)
        if m is None:
            raise ValueError(
                f'Invalid default_run specification "{v}". Allowed arguments: -t/--tasks, -x/--exclude and -u/--upstream-prod'
            )

        include = set()
        exclude = set()
        upstream_prod = None
        for arg in RE_DEFAULT_RUN.findall(v):
            arg = arg.strip()
            if arg.startswith("-t") or arg.startswith("--tasks"):
                include.update(arg.split(" ")[1:])
            elif arg.startswith("-x") or arg.startswith("--exclude"):
                exclude.update(arg.split(" ")[1:])
            elif arg.startswith("-u") or arg.startswith("--upstream-prod"):
                upstream_prod = True
            else:
                raise ValueError('Incorrect option in default_run "{arg}"')

        return {
            "include": include,
            "exclude": exclude,
            "upstream_prod": upstream_prod,
        }


class SettingsYaml(BaseModel):
    class Profile(BaseModel):
        parameters: Optional[Mapping[str, Any]]
        credentials: Mapping[str, str]

        database_prefix: Optional[str]
        database_suffix: Optional[str]
        database_stringify: Optional[str]
        schema_prefix: Optional[str]
        schema_suffix: Optional[str]
        schema_override: Optional[str]
        table_prefix: Optional[str]
        table_suffix: Optional[str]
        table_override: Optional[str]
        from_prod: Optional[Sequence[TableGlob]]
        default_run: Optional[str]

        class Config:
            extra = Extra.forbid
            anystr_lower = True

        @validator("default_run")
        def default_run_validator(cls, v):
            m = RE_DEFAULT_RUN_VAL.match(v)
            if m is None:
                raise ValueError(
                    f'Invalid default_run specification "{v}". Allowed arguments: -t/--tasks, -x/--exclude and -u/--upstream-prod'
                )

            include = set()
            exclude = set()
            upstream_prod = None
            for arg in RE_DEFAULT_RUN.findall(v):
                arg = arg.strip()
                if arg.startswith("-t") or arg.startswith("--tasks"):
                    include.update(arg.split(" ")[1:])
                elif arg.startswith("-x") or arg.startswith("--exclude"):
                    exclude.update(arg.split(" ")[1:])
                elif arg.startswith("-u") or arg.startswith("--upstream-prod"):
                    upstream_prod = True
                else:
                    raise ValueError('Incorrect option in default_run "{arg}"')

            return {
                "include": include,
                "exclude": exclude,
                "upstream_prod": upstream_prod,
            }

    credentials: Mapping[str, Mapping]
    profiles: Mapping[str, Profile]
    default_profile: Optional[str]

    class Config:
        extra = Extra.forbid
        anystr_lower = True

    @validator("profiles")
    def yaml_credentials(cls, v, values):
        if v is None:
            raise ValueError("No profiles defined in settings.yaml.")
        for profile_name, profile in v.items():
            for settings_name in profile.credentials.values():
                if "credentials" not in values:
                    raise ValueError("No credentials defined in settings.yaml.")

                if settings_name not in values["credentials"]:
                    raise ValueError(
                        f'"{settings_name}" in profile "{profile_name}" not declared in credentials.'
                    )

        return v

    @validator("default_profile", always=True)
    def default_profile_exists(cls, v, values):
        if "profiles" not in values:
            # We're running this always so that we can default to the first profile
            # so we cover the case of no profile specified, but the error will be
            # covered by the profiles validation
            return v

        if v is None and len(values["profiles"]) > 1:
            raise ValueError(
                'Can\'t determine default profile. Use "default_profile" to specify it.'
            )
        elif v is None:
            return list(values["profiles"].keys())[0]

        if "profiles" not in values:
            raise ValueError("No profiles defined in settings.yaml.")
        elif v not in values["profiles"].keys():
            raise ValueError(f'default_profile "{v}" not in the profiles map.')

        return v

    def get_profile_info(self, profile_name=None):
        if profile_name is None:
            profile_name = self.default_profile

        if profile_name not in self.profiles:
            raise ValueError(f'Profile "{profile_name}" not in settings.yaml.')

        if profile_name is None:
            raise ValueError("Unknown profile")

        profile_info = self.profiles[profile_name]

        return {
            "parameters": profile_info.parameters,
            "credentials": {
                cred_project_name: self.credentials[cred_settings_name]
                for cred_project_name, cred_settings_name in profile_info.credentials.items()
            },
            "stringify": {
                k: v
                for k, v in {
                    f"{obj_type}_{str_type}": profile_info.dict()[
                        f"{obj_type}_{str_type}"
                    ]
                    for obj_type in ("schema", "table")
                    for str_type in ("prefix", "suffix", "override")
                }.items()
                if v is not None
            },
            "from_prod": [f for f in profile_info.from_prod or list()],
            "default_run": profile_info.default_run
            or {"include": set(), "exclude": set(), "upstream_prod": None},
        }


def read_settings():
    # First read the settings.yaml
    filepath = Path("settings.yaml")
    if filepath.exists():
        try:
            yaml_content = YAML().load(Path(filepath).read_text())
        except MarkedYAMLError as e:
            return Exc(e, filename=filepath)

        try:
            settings_yaml = SettingsYaml(**yaml_content)
        except ValidationError as e:
            return Exc(e, where="settings_reading")
    else:
        settings_yaml = None

    # Process the environment variables
    environment = {
        "parameters": dict(),
        "credentials": dict(),
        "stringify": dict(),
        "from_prod": list(),
    }
    for name, value in os.environ.items():
        matched_name = RE_ENV_VAR_NAME.match(name)
        if matched_name is not None:
            matched_name = matched_name.groupdict()

            if matched_name["type"] is not None:
                environment[matched_name["type"].lower() + "s"][
                    matched_name["name"]
                ] = YAML().load(value)
            elif matched_name["from_prod"] is not None:
                environment["from_prod"] = [v.strip() for v in value.split(",")]
            elif matched_name["default_run"] is not None:
                environment["default_run"] = value
            else:
                environment["stringify"][matched_name["stringify"].lower()] = value

    environment = {k: v for k, v in environment.items() if len(v) > 0}
    try:
        settings_env = Environment(**environment)
    except ValidationError as e:
        return Exc(e, where="settings_reading")

    if len(environment) == 0:
        environment = None

    return Ok({"yaml": settings_yaml, "env": settings_env})


def get_settings(yaml, environment, profile_name=None):
    if profile_name is not None and yaml is None:
        return Err("get_settings", "missing_settings_yaml")
    elif yaml is not None:
        out = yaml.get_profile_info(profile_name)
    else:
        out = {
            "credentials": dict(),
            "parameters": dict(),
            "stringify": dict(),
            "from_prod": list(),
            "default_run": {"include": set(), "exclude": set(), "upstream_prod": None},
        }

    if profile_name is None and environment is not None:
        # When no profile is specified, and there's something in the environment,
        # we try to use environment variables
        if environment.parameters is not None:
            out["parameters"].update(environment.parameters)

        if environment.credentials is not None:
            out["credentials"].update(environment.credentials)

        if environment.stringify is not None:
            out["stringify"].update(
                {k: v for k, v in environment.stringify if v is not None}
            )

        if environment.from_prod is not None:
            out["from_prod"] = environment.from_prod

        if environment.default_run is not None:
            out["default_run"] = environment.default_run

    return Ok(out)


def get_connections(credentials):
    out = dict()
    for name, config in credentials.items():
        try:
            if config is None:
                out[name] = create_dummy(name)
            elif config["type"] == "api":
                out[name] = {k: v for k, v in config.items() if k != "type"}
            else:
                out[name] = create_db(
                    name,
                    name,
                    deepcopy(config),
                )
        except Exception as e:
            return Exc(e)

    return Ok(out)



from pydantic import ValidationError
from ruamel.yaml import YAML
from ruamel.yaml.error import MarkedYAMLError

from ..core.errors import SaynMissingFileError, SaynParsingError


def read_yaml_file(file, Model):
    if not file.exists():
        raise SaynMissingFileError(str(file))

    try:
        parsed = YAML().load(file.read_text())
    except MarkedYAMLError as exc:
        raise SaynParsingError(
            "yaml_parsing",
            [
                {
                    "file_name": str(file),
                    "line": exc.problem_mark.line,
                    "column": exc.problem_mark.column,
                    "message": exc.problem,
                    "snippet": exc.problem_mark.get_snippet(),
                }
            ],
        )

    try:
        return Model(**parsed)
    except ValidationError as exc:
        errors = list()
        for error in exc.errors():
            tmp_parsed = parsed
            for k in error["loc"][:-1]:
                tmp_parsed = tmp_parsed[k]
            lc = tmp_parsed.lc.key(error["loc"][-1])

            errors.append(
                {
                    "file_name": str(file),
                    "loc": error["loc"],
                    "line": lc[0] + 1,
                    "column": lc[1] + 1,
                    "message": error["msg"],
                }
            )

        raise SaynParsingError("data_validation", errors)



from .misc import reverse_dict


def plot_dag(tasks, folder=None, file_name=None):
    """Uses graphviz to plot the dag
    It requires the graphviz python package (pip install graphviz) and an installation of graphviz
    (eg: brew install graphviz)
    """

    task_list = list(tasks.keys())
    tasks = reverse_dict(tasks)

    try:
        from graphviz import Digraph
    except:
        raise ImportError("`sayn dag-image` requires graphviz installed")

    dot = Digraph(comment="SAYN", graph_attr={"splines": "ortho", "nodesep": "0.8"})
    for n in task_list:
        dot.node(n, style="rounded", shape="box")
    for a, l in tasks.items():
        if a in task_list:
            for b in l:
                if b in task_list:
                    dot.edge(a, b)

    if file_name is not None:
        dot.render(
            directory=folder,
            filename=file_name,
            cleanup=True,
            view=False,
            format="png",
        )



from .logger import Logger
from .log_formatter import LogFormatter


class ConsoleLogger(Logger):
    fmt = LogFormatter(use_colour=True, output_ts=True)
    is_debug = True

    def __init__(self, debug=True):
        self.is_debug = debug

    def print(self, s=None):
        if s is None:
            print()
        else:
            prefix = "  " * self.current_indent
            s = s["message"]
            if isinstance(s, str):
                s = [s]
            elif not isinstance(s, list):
                raise ValueError("error in logging print")

            print(f"{prefix}{s[0]}")
            for e in s[1:]:
                for l in e.split("\n"):
                    print(f"{prefix}  {l}")



from datetime import date, timedelta
import sys

import click

from .utils.graphviz import plot_dag
from .logging import ConsoleLogger, FancyLogger, FileLogger
from .scaffolding.init_project import sayn_init
from .core.app import App, Command
from .tasks.task import TaskStatus

yesterday = date.today() - timedelta(days=1)


class CliApp(App):
    def __init__(
        self,
        command,
        debug=False,
        include=None,
        exclude=None,
        upstream_prod=False,
        profile=None,
        full_load=False,
        start_dt=None,
        end_dt=None,
    ):
        super().__init__()

        # STARTING APP: register loggers and set cli arguments in the App object
        self.run_arguments.command = command

        if debug:
            self.run_arguments.debug = debug
        else:
            self.tracker.remove_logger(ConsoleLogger)
            self.tracker.register_logger(FancyLogger())

        self.tracker.register_logger(
            FileLogger(
                self.run_arguments.folders.logs,
                format=f"{self.run_id}|" + "%(asctime)s|%(levelname)s|%(message)s",
            )
        )

        if start_dt is not None:
            self.run_arguments.dates_specified = True
            self.run_arguments.start_dt = start_dt.date()
        else:
            self.run_arguments.start_dt = yesterday

        if end_dt is not None:
            self.run_arguments.dates_specified = True
            self.run_arguments.end_dt = end_dt.date()
        else:
            end_dt = yesterday
            self.run_arguments.end_dt = yesterday

        self.run_arguments.profile = profile
        self.run_arguments.full_load = full_load

        if include is not None:
            self.run_arguments.include = set(include)

        if exclude is not None:
            self.run_arguments.exclude = set(exclude)

        if upstream_prod is not None:
            self.run_arguments.upstream_prod = upstream_prod

        self.start_app()


class ChainOption(click.Option):
    def __init__(self, *args, **kwargs):
        self.save_other_options = kwargs.pop("save_other_options", True)
        nargs = kwargs.pop("nargs", -1)
        assert nargs == -1, "nargs, if set, must be -1 not {}".format(nargs)
        super(ChainOption, self).__init__(*args, **kwargs)
        self._previous_parser_process = None
        self._eat_all_parser = None

    def add_to_parser(self, parser, ctx):
        def parser_process(value, state):
            # method to hook to the parser.process
            done = False
            if self.save_other_options:
                # grab everything up to the next option
                while state.rargs and not done:
                    for prefix in self._eat_all_parser.prefixes:
                        if state.rargs[0].startswith(prefix):
                            done = True
                    if not done:
                        value += f" {state.rargs.pop(0)}"
            else:
                # grab everything remaining
                value += state.rargs
                state.rargs[:] = []

            # call the actual process
            self._previous_parser_process(value, state)

        retval = super(ChainOption, self).add_to_parser(parser, ctx)
        for name in self.opts:
            our_parser = parser._long_opt.get(name) or parser._short_opt.get(name)
            if our_parser:
                self._eat_all_parser = our_parser
                self._previous_parser_process = our_parser.process
                our_parser.process = parser_process
                break
        return retval


# Click arguments

click_debug = click.option(
    "--debug", "-d", is_flag=True, default=False, help="Include debug messages"
)


def click_filter(func):
    func = click.option(
        "--tasks",
        "-t",
        multiple=True,
        cls=ChainOption,
        help="Task query to INCLUDE in the execution: [+]task_name[+], group:group_name, tag:tag_name",
        default=list(),
    )(func)
    func = click.option(
        "--exclude",
        "-x",
        multiple=True,
        cls=ChainOption,
        help="Task query to EXCLUDE in the execution: [+]task_name[+], group:group_name, tag:tag_name",
        default=list(),
    )(func)
    func = click.option(
        "--upstream-prod",
        "-u",
        is_flag=True,
        default=False,
        help="For database objects produced by tasks not being executed, use the production object",
    )(func)
    return func


def click_incremental(func):
    func = click.option(
        "--full-load", "-f", is_flag=True, default=False, help="Do a full load"
    )(func)
    func = click.option(
        "--start-dt",
        "-s",
        type=click.DateTime(formats=["%Y-%m-%d"]),
        default=None,
        help="For incremental loads, the start date",
    )(func)
    func = click.option(
        "--end-dt",
        "-e",
        type=click.DateTime(formats=["%Y-%m-%d"]),
        default=None,
        help="For incremental loads, the end date",
    )(func)
    return func


def click_run_options(func):
    func = click_debug(func)
    func = click.option("--profile", "-p", help="Profile from settings to use")(func)
    func = click_incremental(func)
    func = click_filter(func)
    return func


@click.group(help="SAYN management tool.")
def cli():
    pass


# Click commands


@cli.command(help="Initialise a SAYN project in working directory.")
@click.argument("sayn_project_name")
def init(sayn_project_name):
    sayn_init(sayn_project_name)


@cli.command(help="Compile sql tasks.")
@click_run_options
def compile(debug, tasks, exclude, upstream_prod, profile, full_load, start_dt, end_dt):

    tasks = [i for t in tasks for i in t.strip().split(" ")]
    exclude = [i for t in exclude for i in t.strip().split(" ")]
    app = CliApp(
        Command.COMPILE,
        debug,
        tasks,
        exclude,
        upstream_prod,
        profile,
        full_load,
        start_dt,
        end_dt,
    )

    app.compile()
    if any([t.status == TaskStatus.FAILED for _, t in app.tasks.items()]):
        sys.exit(-1)
    else:
        sys.exit()


@cli.command(help="Run SAYN tasks.")
@click_run_options
def run(debug, tasks, exclude, upstream_prod, profile, full_load, start_dt, end_dt):

    tasks = [i for t in tasks for i in t.strip().split(" ")]
    exclude = [i for t in exclude for i in t.strip().split(" ")]
    app = CliApp(
        Command.RUN,
        debug,
        tasks,
        exclude,
        upstream_prod,
        profile,
        full_load,
        start_dt,
        end_dt,
    )

    app.run()
    if any([t.status == TaskStatus.FAILED for _, t in app.tasks.items()]):
        sys.exit(-1)
    else:
        sys.exit()


@cli.command(help="Test SAYN tasks.")
@click_run_options
def test(debug, tasks, exclude, upstream_prod, profile, full_load, start_dt, end_dt):

    tasks = [i for t in tasks for i in t.strip().split(" ")]
    exclude = [i for t in exclude for i in t.strip().split(" ")]
    app = CliApp(
        Command.TEST,
        debug,
        tasks,
        exclude,
        upstream_prod,
        profile,
        full_load,
        start_dt,
        end_dt,
    )

    app.test()
    if any([t.status == TaskStatus.FAILED for _, t in app.tasks.items()]):
        sys.exit(-1)
    else:
        sys.exit()


@cli.command(help="Generate DAG image.")
@click_debug
@click_filter
def dag_image(debug, tasks, exclude, upstream_prod):
    def handle_error():
        print("Errors detected in project. Run `sayn compile` to see the errors")
        sys.exit(-1)

    app = App()
    app.start_app()

    plot_dag(app.dag, "images", "dag")

    print("Dag image created in `images/dag.png`")



from pathlib import Path
from datetime import datetime
import subprocess
from typing import List, Optional

from .task_event_tracker import TaskEventTracker

from sayn import __version__ as sayn_version


class EventTracker:
    loggers = list()
    run_id = None
    current_stage = None
    current_stage_start_ts: Optional[datetime]
    tasks: List
    task_trackers: List[TaskEventTracker]
    current_task = None
    current_task_n = 0
    sayn_version = sayn_version
    project_git_commit = None
    project_name = Path(".").absolute().name

    def __init__(self, run_id):
        self.run_id = run_id
        self.tasks = list()
        try:
            self.project_git_commit = (
                subprocess.check_output(
                    ["git", "rev-parse", "HEAD"], stderr=subprocess.STDOUT
                )
                .split()[0]
                .decode("utf-8")
            )
        except:
            # If git is not available, we simply don't report the commit
            pass

    def register_logger(self, logger):
        self.loggers.append(logger)

    def remove_logger(self, logger_type):
        self.loggers = [l for l in self.loggers if not isinstance(l, logger_type)]

    def start_stage(self, stage, **details):
        self.current_stage = stage
        self.current_stage_start_ts = datetime.now()
        self.report_event(context="app", event="start_stage", stage=stage, **details)

    def finish_current_stage(self, **details):
        if self.current_stage_start_ts is None:
            self.current_stage_start_ts = datetime.now()

        duration = datetime.now() - self.current_stage_start_ts
        self.report_event(
            context="app",
            event="finish_stage",
            stage=self.current_stage,
            duration=duration,
            **details
        )
        self.current_stage = None
        self.current_stage_start_ts = None

    def set_tasks(self, tasks):
        self.tasks = tasks

    def get_task_tracker(self, task_name):
        if task_name in self.tasks:
            task_order = self.tasks.index(task_name) + 1
        else:
            task_order = None

        return TaskEventTracker(self, task_name, task_order)

    def report_event(self, **event):
        if "context" not in event:
            event["context"] = "app"

        elif event["context"] == "task":
            event["total_tasks"] = len(self.tasks)

        if "event" not in event:
            event["event"] = "unknown"

        event.update(
            dict(
                run_id=self.run_id,
                stage=self.current_stage,
                sayn_version=self.sayn_version,
                project_git_commit=self.project_git_commit,
                project_name=self.project_name,
                ts=datetime.now(),
            )
        )

        for logger in self.loggers:
            logger.report_event(**event)



from copy import deepcopy
from pathlib import Path
from typing import Any, List, Mapping, Optional

from pydantic import BaseModel, Field, validator, Extra
from sayn.tasks.python import DecoratorTaskWrapper

from ..utils.compiler import TaskJinjaEnv
from ..utils.misc import merge_dicts, merge_dict_list
from ..utils.dag import upstream, topological_sort
from ..utils.yaml import read_yaml_file
from .errors import Err, Ok


class Project(BaseModel):
    required_credentials: List[str]
    default_db: Optional[str]
    parameters: Optional[Mapping[str, Any]]
    presets: Optional[Mapping[str, Mapping[str, Any]]]
    autogroups: Mapping[str, Any] = Field(dict(), alias="groups")

    database_prefix: Optional[str]
    database_suffix: Optional[str]
    database_override: Optional[str]
    schema_prefix: Optional[str]
    schema_suffix: Optional[str]
    schema_override: Optional[str]
    table_prefix: Optional[str]
    table_suffix: Optional[str]
    table_override: Optional[str]

    class Config:
        extra = Extra.forbid
        anystr_lower = True

    @validator("required_credentials")
    def required_credentials_are_unique(cls, v):
        if len(set(v)) != len(v):
            raise ValueError("Duplicate values found in required_credentials")
        return v

    @validator("default_db", always=True)
    def default_db_exists(cls, v, values):
        if "required_credentials" not in values:
            return v

        if v is None and len(values["required_credentials"]) == 1:
            return values["required_credentials"][0]
        elif v is None:
            raise ValueError("Missing default_db in project.yaml")
        elif v not in values["required_credentials"]:
            raise ValueError(f'default_db value "{v}" not in required_credentials')

        return v

    @validator(
        *[
            f"{o}_{t}"
            for o in ("table", "schema")
            for t in ("prefix", "suffix", "override")
        ],
        allow_reuse=True,
    )
    def non_empty_strings(cls, v):
        if len(v.strip()) != 0:
            return v.strip()


def read_project(project_root=Path(".")):
    return read_yaml_file(project_root / Path("project.yaml"), Project)


class TaskGroupFile(BaseModel):
    presets: Optional[Mapping[str, Mapping[str, Any]]]
    tasks: Optional[Mapping[str, Mapping[str, Any]]]
    tests: Optional[Mapping[str, Mapping[str, Any]]]

    class Config:
        extra = Extra.forbid
        anystr_lower = True


def read_groups(project_root=Path(".")):
    task_folder = project_root / "tasks"
    if task_folder.exists() and task_folder.is_dir():
        out = dict()
        for file in task_folder.glob("*.yaml"):
            name = str(file.relative_to(task_folder))[:-5]
            out[name] = read_yaml_file(file, TaskGroupFile)

        return out
    else:
        return dict()


###############################
# Task related config functions
###############################


def get_presets(global_presets, groups):
    """Returns a dictionary of presets merged with the referenced preset
    Presets define a direct acyclic graph by including the `preset` property, so
    this function validates that there are no cycles and that all referenced presets
    are defined.
    In the output, preset names are prefixed with `sayn_global:` or `group:` so that we can
    merge all presets in the project in the same dictionary.
    Args:
      global_presets (dict): dictionary containing the presets defined in project.yaml
      groups (sayn.app.config.TaskGroup): a list of task groups from the tasks/ folder
    """
    # 1. Construct a dictionary of presets so we can attach that info to the tasks
    presets_info = {
        f"sayn_global:{k}": {kk: vv for kk, vv in v.items() if kk != "preset"}
        for k, v in global_presets.items()
    }

    # 1.1. We start with the global presets defined in project.yaml
    presets_project = {
        k: [f"sayn_global:{v}"] if v is not None else []
        for k, v in {
            f"sayn_global:{name}": preset.get("preset")
            for name, preset in global_presets.items()
        }.items()
    }

    # 1.2. Then we add the presets defined in the task groups
    for group_name, group in groups.items():
        if group.presets is not None:
            presets_info.update(
                {
                    f"{group_name}:{k}": {
                        kk: vv for kk, vv in v.items() if kk != "preset"
                    }
                    for k, v in group.presets.items()
                }
            )

        if group.presets is None:
            group_presets = dict()
        else:
            group_presets = {
                name: preset.get("preset") for name, preset in group.presets.items()
            }

        # Check if the preset referenced is defined in the task group, otherwise, point at the
        # global task group
        group_presets = {
            f"{group_name}:{k}": [
                f"{group_name}:{v}"
                if v in group_presets and v != k
                else f"sayn_global:{v}"
            ]
            if v is not None
            else []
            for k, v in group_presets.items()
        }
        presets_project.update(group_presets)

    # 1.3. The preset references represent a dag that we need to validate, ensuring
    #      there are no cycles and that all references exists
    result = topological_sort(presets_project)
    if result.is_err:
        return result
    else:
        topo_sort = result.value

    # 1.4. Merge the presets with the reference preset, so that we have 1 dictionary
    #      per preset a task could reference
    presets = {
        name: merge_dict_list(
            [presets_info[p] for p in upstream(presets_project, name).value]
            + [presets_info[name]]
        )
        for name in topo_sort
    }

    return Ok(presets)


def get_task_dict(task, task_name, group_name, presets):
    """Returns a single task merged with the referenced preset
    Args:
      task (dict): a dictionary with the task information
      task_name (str): the name of the task
      group_name (str): the name of the group it appeared on
      presets (dict): a dictionary of merged presets returned by get_presets
    """
    if "preset" in task:
        preset_name = task["preset"]
        preset = presets.get(
            f"{group_name}:{preset_name}",
            presets.get(f"sayn_global:{preset_name}"),
        )
        if preset is None:
            return Err(
                "get_task_dict",
                "missing_preset",
                group=group_name,
                task=task_name,
                preset=preset_name,
            )
        task = merge_dicts(preset, task)

    return Ok(dict(task, name=task_name, group=group_name))


def get_tasks_dict(
    global_presets, groups, autogroups, sql_folder, compiler, python_loader
):
    """Returns a dictionary with the task definition with the preset information merged
    Args:
      global_presets (dict): a dictionary with the presets as defined in project.yaml
      groups (sayn.common.config.TaskGroup): a list of task groups from the tasks/ folder
    """
    result = get_presets(global_presets, groups)
    if result.is_err:
        return result
    else:
        presets = result.value

    errors = dict()
    tasks = dict()
    tests = dict()

    for group_name, group in groups.items():
        if group.tasks is not None:
            for task_name, task in group.tasks.items():
                if task_name in tasks:
                    return Err(
                        "dag",
                        "duplicate_task",
                        task=task_name,
                        groups=(group_name, tasks[task_name]["group"]),
                    )
                result = get_task_dict(task, task_name, group_name, presets)
                if result.is_ok and "type" in result.value:
                    tasks[task_name] = result.value
                    if tasks[task_name].get("type") == "test":
                        return Err(
                            "dag",
                            "test in tasks",
                            task=task_name,
                        )
                else:
                    if "type" in result.value:
                        errors[task_name] = result.error
                    else:
                        errors[task_name] = Err(
                            "get_task_dict",
                            "no_type",
                            group=group_name,
                            task=task_name,
                        )

        if group.tests is not None:
            for test_name, test in group.tests.items():
                if "type" in test.keys():
                    return Err(
                        "dag",
                        "Tests have no types",
                        task=test_name,
                    )

                test["type"] = "test"

                if test_name in tests:
                    return Err(
                        "dag",
                        "duplicate_task",
                        task=test_name,
                        groups=(group_name, tasks[test_name]["group"]),
                    )

                result = get_task_dict(test, test_name, group_name, presets)
                if result.is_ok:
                    tests[test_name] = result.value
                else:
                    errors[test_name] = result.error

    for group_name, group in autogroups.items():
        group_definition = {"group": group_name}
        if "preset" in group:
            preset_name = f"sayn_global:{group['preset']}"
            if preset_name not in presets:
                return Err("dag", "missing_preset", preset_name=preset_name)
            group_definition.update(deepcopy(presets[preset_name]))

        group_definition.update(group)

        if group_definition.get("type") in ("sql", "autosql"):
            if "file_name" not in group_definition:
                return Err(
                    "dag",
                    "missing_file_name",
                    group=group_name,
                    type=group_definition.get("type"),
                )

            file_glob = compiler.compile(
                group_definition["file_name"], task=TaskJinjaEnv(group=group_name)
            )
            found_file = False

            for file in Path(sql_folder).glob(file_glob):
                found_file = True
                task_name = file.stem
                if task_name in tasks:
                    return Err(
                        "dag",
                        "duplicate_task",
                        task_name=task_name,
                        groups=(group_name, tasks[task_name]["group"]),
                    )

                task = deepcopy(group_definition)
                task["file_name"] = str(file.relative_to(sql_folder))

                result = get_task_dict(task, task_name, group_name, presets)
                if result.is_ok:
                    tasks[task_name] = result.value
                else:
                    errors[task_name] = result.error

                tasks[task_name] = task

            if not found_file:
                return Err("dag", "empty_group", group=group_name)

        elif group_definition.get("type") == "python":
            group_definition["group"] = group_name
            group_definition["type"] = "python_module"
            if "module" not in group_definition:
                return Err("task", "missing_module", group=group_name)
            module_path = compiler.compile(
                group_definition.pop("module", None),
                task=TaskJinjaEnv(group=group_name),
            )
            result = python_loader.get_objects(
                "python_tasks", module_path, DecoratorTaskWrapper
            )
            if result.is_err:
                return result
            else:
                objects = result.value

            for task_obj in objects:
                task_name = task_obj.func.__name__
                task = deepcopy(group_definition)
                task["name"] = task_name
                task["task_class"] = task_obj

                result = get_task_dict(task, task_name, group_name, presets)
                if result.is_ok:
                    tasks[task_name] = result.value
                else:
                    errors[task_name] = result.error

        elif group_definition.get("type") == "test":
            if "file_name" not in group_definition:
                return Err(
                    "dag",
                    "missing_file_name",
                    group=group_name,
                    type=group_definition.get("type"),
                )
            file_glob = compiler.compile(
                group_definition["file_name"], task=TaskJinjaEnv(group=group_name)
            )

            found_file = False
            for file in Path(sql_folder).glob(file_glob):
                found_file = True
                task_name = file.stem
                if task_name in tasks:
                    return Err(
                        "dag",
                        "duplicate_task",
                        task_name=task_name,
                        groups=(group_name, tasks[task_name]["group"]),
                    )
                task = deepcopy(group_definition)
                task["file_name"] = str(file.relative_to(sql_folder))

                result = get_task_dict(task, task_name, group_name, presets)

                if result.is_ok:
                    tests[task_name] = result.value
                else:
                    errors[task_name] = result.error

                tests[task_name] = task

        # elif group_definition.get("type") == "python":
        #     group_definition["group"] = group_name
        #     group_tasks = group_definition.pop("tasks", dict())
        #     if not isinstance(group_tasks, dict):
        #         return Err(
        #             "dag",
        #             "wrong_yaml_type",
        #             group=group_name,
        #             type=type(group_definition.get("tasks")),
        #         )

        #     for task_name, task_def in group_tasks.items():
        #         if task_name in tasks:
        #             return Err(
        #                 "dag",
        #                 "duplicate_task",
        #                 task_name=task_name,
        #                 groups=(group_name, tasks[task_name]["group"]),
        #             )

        #         if task_def is not None:
        #             if not isinstance(task_def, dict):
        #                 return Err(
        #                     "dag",
        #                     "wrong_yaml_type",
        #                     group=group_name,
        #                     type=type(task_def),
        #                 )

        #             task = deepcopy(group_definition)
        #             task = merge_dicts(task, task_def)
        #         else:
        #             task = deepcopy(group_definition)

        #         result = get_task_dict(task, task_name, group_name, presets)
        #         if result.is_ok:
        #             tasks[task_name] = result.value
        #         else:
        #             errors[task_name] = result.error

        else:
            return Err(
                "dag",
                "wrong_autogroup_type",
                group=group_name,
                type=group_definition.get("type"),
            )

    for t in tests:
        if t in tasks.keys():
            return Err("dag", "Duplicate tests in tasks", task=t)

    if len(errors) > 0:
        return Err("get_tasks_dict", "task_parsing_error", errors=errors)
    else:
        return Ok(merge_dicts(tasks, tests))



from .log_formatter import LogFormatter


class Logger:
    fmt = LogFormatter()
    current_indent = 0

    def unhandled(self, event, context, stage, details):
        self.print(self.fmt.unhandled(event, context, stage, details))

    def message(self, level, message, details):
        self.print(self.fmt.message(level, message, details))

    # App context

    def app_start(self, details):
        self.print(self.fmt.app_start(details))
        self.print()

    def app_finish(self, details):
        self.print(self.fmt.app_finish(details))

    def app_stage_start(self, stage, details):
        self.print(self.fmt.app_stage_start(stage, details))
        self.current_indent += 1

    def app_stage_finish(self, stage, details):
        self.current_indent -= 1
        self.print(self.fmt.app_stage_finish(stage, details))
        self.print()

    # Task context

    def task_stage_start(self, stage, task, task_order, total_tasks, details):
        self.print(
            self.fmt.task_stage_start(stage, task, task_order, total_tasks, details)
        )
        self.current_indent += 1

    def task_stage_finish(self, stage, task, task_order, total_tasks, details):
        self.current_indent -= 1
        self.print(
            self.fmt.task_stage_finish(stage, task, task_order, total_tasks, details)
        )
        if stage in ("run", "compile", "test"):
            self.print()

    def task_set_steps(self, details):
        self.print(self.fmt.task_set_steps(details))

    def task_step_start(self, stage, task, step, step_order, total_steps, details):
        self.print(
            self.fmt.task_step_start(
                stage, task, step, step_order, total_steps, details
            )
        )
        self.current_indent += 1

    def task_step_finish(self, stage, task, step, step_order, total_steps, details):
        self.current_indent -= 1
        self.print(
            self.fmt.task_step_finish(
                stage, task, step, step_order, total_steps, details
            )
        )

    def report_event(self, context, event, stage, **details):
        if event == "message":
            self.message(details["level"], details["message"], details)

        elif context == "app":
            if event == "start_app":
                self.app_start(details)

            elif event == "finish_app":
                self.app_finish(details)

            elif event == "start_stage":
                self.app_stage_start(stage, details)

            elif event == "finish_stage":
                self.app_stage_finish(stage, details)

            else:
                self.unhandled(event, context, stage, details)

        elif context == "task":
            task = details["task"]
            task_order = details["task_order"]
            total_tasks = details["total_tasks"]

            if event == "set_run_steps":
                self.task_set_steps(details)

            elif event == "start_stage":
                self.task_stage_start(stage, task, task_order, total_tasks, details)

            elif event == "finish_stage":
                self.task_stage_finish(stage, task, task_order, total_tasks, details)

            elif event == "start_step":
                self.task_step_start(
                    stage,
                    task,
                    details["step"],
                    details["step_order"],
                    details["total_steps"],
                    details,
                )

            elif event == "finish_step":
                self.task_step_finish(
                    stage,
                    task,
                    details["step"],
                    details["step_order"],
                    details["total_steps"],
                    details,
                )

            else:
                self.unhandled(event, context, stage, details)

        else:
            self.unhandled(event, context, stage, details)

    def print(self, s=None):
        raise NotImplementedError("Logger requires print method")



from contextlib import contextmanager
from enum import Enum
from pathlib import Path
from typing import Set, Dict, Any

from colorama import Fore, Style

from ..core.errors import Err, Ok
from ..logging.task_event_tracker import TaskEventTracker
from ..utils.compiler import Compiler


class TaskStatus(Enum):
    NOT_IN_QUERY = "not_in_query"

    CONFIGURING = "config"
    READY_FOR_SETUP = "ready_for_setup"

    SETTING_UP = "setting_up"
    READY = "ready"
    SETUP_FAILED = "setup_failed"

    EXECUTING = "executing"
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    SKIPPED = "skipped"

    UNKNOWN = "unknown"


class Task:
    """
    Base class for tasks in SAYN.

    Attributes:
        name (str): Name of the task as defined in the task group.
        group (str): Name of the task group where the task was defined.
        run_arguments (dict): Dictionary containing the values for the arguments specified in the cli.
        task_parameters (dict): Provides access to the parameters specified in the task.
        project_parameters (dict): Provides access to the global parameters of the project.
        parameters (dict): Convinience property joining project and task parameters.
        connections (dict): Dictionary of connections specified for the project.
        tracker (sayn.logging.TaskEventTracker): Message tracker for the current task.
        jinja_env (jinja2.Environment): Jinja environment for this task. The environment comes pre-populated with the parameter values relevant to the task.
    """

    name: str
    group: str
    tags: Set[str]
    run_arguments: Dict[str, Any]
    task_parameters: Dict[str, Any]
    project_parameters: Dict[str, Any]

    _default_db: str
    connections = dict()
    _tracker: TaskEventTracker

    compiler: Compiler

    _has_tests = False
    _needs_recompile = False

    # Handy properties
    @property
    def parameters(self):
        return {**self.project_parameters, **self.task_parameters}

    @property
    def default_db(self):
        return self.connections[self._default_db]

    # Making it backwards compatible
    @property
    def logger(self):
        return self._tracker

    @property
    def needs_recompile(self):
        return self._needs_recompile

    # Task lifetime methods

    def config(self, **config):
        raise NotImplementedError(
            "SAYN task", self.__class__.__name__, "setup", str(config)
        )

    def setup(self):
        raise NotImplementedError("SAYN task", self.__class__.__name__, "setup")

    def run(self):
        raise NotImplementedError("SAYN task", self.__class__.__name__, "run")

    def test(self):
        raise NotImplementedError("SAYN task", self.__class__.__name__, "test")

    def compile(self):
        raise NotImplementedError("SAYN task", self.__class__.__name__, "compile")

    # Status methods

    def __init__(
        self,
        name,
        group,
        tracker,
        run_arguments,
        task_parameters,
        project_parameters,
        default_db,
        connections,
        compiler,
        src,
        out,
    ):
        self.name = name
        self.group = group
        self._tracker = tracker
        self.run_arguments = run_arguments
        self.task_parameters = task_parameters
        self.project_parameters = project_parameters
        self._default_db = default_db
        self.connections = connections
        self.compiler = compiler
        self.src = src
        self.out = out

        # This is a dictionary of configuration that can be passed from the runner to the wrapper
        self._config_input = {
            "sources": set(),
            "outputs": set(),
            "parents": set(),
            "tags": set(),
            "on_fail": None,
            "task_name": None,
        }

    def ready(self):
        """(Deprecated: use `success` instead) Returned on successful execution."""
        return Ok()

    def success(self):
        """Returned on successful execution."""
        return Ok()

    def fail(self, msg=None):
        """Returned on failure in any stage."""
        if msg is None:
            msg = 'Unknown error. Use `self.fail("Error message")` in python tasks for more details.'
        return Err("tasks", "task_fail", message=msg)

    # Logging methods

    def set_run_steps(self, steps):
        """Sets the run steps for the task, allowing the CLI to indicate task execution progress."""
        self._tracker.set_run_steps(steps)

    def add_run_steps(self, steps):
        """Adds new steps to the list of run steps for the task, allowing the CLI to indicate task execution progress."""
        self._tracker.add_run_steps(steps)

    def start_step(self, step):
        """Specifies the start of a task step"""
        self._tracker.start_step(step)

    def finish_current_step(self, result=Ok()):
        """Specifies the end of the current step"""
        self._tracker.finish_current_step(result)

    def debug(self, message, details=None):
        """Print a debug message when executing sayn in debug mode (`sayn run -d`)"""
        if details is None:
            details = dict()
        self._tracker.debug(message, **details)

    def info(self, message, details=None):
        """Prints an info message."""
        if details is None:
            details = dict()
        self._tracker.info(message, **details)

    def warning(self, message, details=None):
        """Prints a warning message which will be persisted on the screen after the task concludes execution."""
        if details is None:
            details = dict()
        self._tracker.warning(message, **details)

    def error(self, message, details=None):
        """Prints an error message which will be persisted on the screen after the task concludes execution.

        Executing this method doesn't abort the task or changes the task status. Use `return self.fail` for that instead.

        Args:
          message (str): An optinal error message to print to the screen.

        """
        if details is None:
            details = dict()
        self._tracker.error(message, **details)

    @contextmanager
    def step(self, step):
        """Step context

        Usage:
        ```python
        with self.step('Generate Data'):
            data = generate_data()
        ```

        Args:
          step (str): name of the step being executed.
        """
        self._tracker.start_step(step)
        yield
        self._tracker.finish_current_step()

    # Test methods

    def get_test_breakdown(self, breakdown: list) -> list:
        """Return list breakdown of: test status, test type and column.

        Usage:
        ```python
        breakdown = self.get_test_breakdown(self.test_breakdown)
        ```

        Args:
          breakdown (list): test_breakdown list given by db method _construct_tests
        """
        data = []
        for brk in breakdown:
            if not brk["execute"]:
                data.append(
                    ["SKIPPED", brk["type"], brk["column"], brk["allowed_values"]]
                )
            else:
                data.append(
                    ["EXECUTED", brk["type"], brk["column"], brk["allowed_values"]]
                )

        return data

    def test_sucessful(self, breakdown: list) -> Ok:
        """CLI outputs on successful test execution.

        Args:
          breakdown (list): output of get_test_breakdown class method.
        """
        skipped = [brk for brk in breakdown if brk[0] == "SKIPPED"]
        executed = [brk for brk in breakdown if brk[0] == "EXECUTED"]

        if skipped:
            self.info(
                f"{Fore.GREEN}{len(skipped)} Column test(s) {Style.BRIGHT}SKIPPED{Style.NORMAL}"
            )
        self.info(
            f"{Fore.GREEN}{len(executed)} Column test(s) {Style.BRIGHT}EXECUTED{Style.NORMAL}, {len(executed)} succeeded."
        )

        return self.success()

    def test_failure(self, breakdown: list, result: dict, run_argument: str) -> tuple:
        """CLI outputs on failed test execution.

        Args:
          breakdown (list): output of get_test_breakdown class method.
          result (dict): output of test query.
          run_argument (str): "debug" entry in class run_arguments.
        """
        skipped = []
        executed = []
        failed = []
        for brk in breakdown:
            if (sum(brk[1] != res["type"] for res in result) == len(result)) or (
                sum(brk[2] != res["col"] for res in result) == len(result)
            ):
                if brk[0] == "SKIPPED":
                    skipped.append(brk)
                if brk[0] == "EXECUTED":
                    executed.append(brk)
            else:
                failed.append(brk)
        if run_argument:

            fl_info = [f"{Fore.RED}FAILED: "]
            for info in failed:
                count = sum(
                    [
                        item["cnt"]
                        for item in result
                        if (item["type"] == info[1] and item["col"] == info[2])
                    ]
                )
                fl_info.append(
                    f"{Fore.RED}{Style.BRIGHT}{info[1]} test{Style.NORMAL} on {Style.BRIGHT}{info[2]} FAILED{Style.NORMAL}. {count} offending records."
                )
            if skipped:
                self.info(
                    f"{Fore.GREEN}{len(skipped)} Column test(s) {Style.BRIGHT}SKIPPED{Style.NORMAL}"
                )
            self.info(
                f"{Fore.GREEN}{len(executed)+len(failed)} Column test(s) {Style.BRIGHT}EXECUTED{Style.NORMAL}, {len(executed)} succeeded."
            )
            for err in fl_info:
                self.info(err)

            errinfo = f"Test Failed. You can find the compiled test query at compile/{self.group}/{self.name}_test.sql. You can find queries to retrieve the problematic values at compile/{self.group}/{self.name}_test_problematic_values.sql"

            return (self.fail(errinfo), failed)
        else:
            summary = f"{len(executed)+len(failed)} Column tests were ran, {len(executed)} succeeded, "
            if skipped:
                summary += f", {len(skipped)} were skipped, "
            summary += f"{len(failed)} failed."
            self.warning(summary)
            errout = ", ".join(list(set([res["type"] for res in result])))

            return (self.fail(f"Failed test types: {errout}"), failed)

    # Jinja methods

    def compile_obj(self, obj, **params):
        """Compiles the object into a string using the task jinja environment.

        Args:
          obj (str/Path/Template): The object to compile. If the object is not a Jinja template object, `self.get_template` will be called first.
          params (dict): An optional dictionary of additional values to use for compilation.
              Note: Project and task parameters values are already set in the environment, so there's no need to pass them on
        """
        return self.compiler.compile(obj, **params)

    def write_compilation_output(self, content, suffix=None, extension="sql"):
        """Writes text content into the compilation folder

        The file will be stored in a folder with the name of the group the task
        belongs to and the name of the file will be that of the task name.
        """
        path = Path(
            self.run_arguments["folders"]["compile"],
            self.group,
            Path(f"{self.name}{'_'+suffix if suffix is not None else ''}.{extension}"),
        )

        # Ensure the path exists and it's empty
        path.parent.mkdir(parents=True, exist_ok=True)
        if path.exists():
            path.unlink()

        path.write_text(str(content))



from .misc import reverse_dict_inclusive

from ..core.errors import Err, Ok


def _has_missing_parents(dag):
    missing = {}
    for task, parents in dag.items():
        missing_parents = [p for p in parents if p not in dag]
        if len(missing_parents) > 0:
            missing[task] = missing_parents
    if len(missing) > 0:
        return Err("dag", "missing_parents", missing=missing)
    return Ok(False)


def _is_cyclic_helper(dag, node, visited, stack, breadcrumbs):
    visited[node] = True
    stack[node] = True

    for parent in dag[node]:
        if parent == node:
            return Err("dag", "cycle_error", path=[node, parent])
        elif not visited[parent]:
            result = _is_cyclic_helper(
                dag, parent, visited, stack, breadcrumbs + [parent]
            )
            if result.is_err:
                return result
        elif stack[parent]:
            return Err("dag", "cycle_error", path=breadcrumbs + [parent])

    stack[node] = False

    return Ok()


def _is_cyclic(dag):
    visited = {n: False for n in dag.keys()}
    stack = {n: False for n in dag.keys()}
    for node in dag.keys():
        if not visited[node]:
            result = _is_cyclic_helper(dag, node, visited, stack, [node])
            if result.is_err:
                return result

    return Ok(True)


def dag_is_valid(dag):
    result = _has_missing_parents(dag)
    if result.is_err:
        return result
    result = _is_cyclic(dag)
    if result.is_err:
        return result

    return Ok()


# DAG -> Sorted list
def topological_sort(dag):
    if len(dag) == 0:
        return Ok(list())
    result = dag_is_valid(dag)
    if result.is_err:
        return result
    topo_sorted = []
    pending = list(dag.keys())
    current = 0
    while True:
        if len(set(dag[pending[current]]) - set(topo_sorted)) == 0:
            topo_sorted.append(pending.pop(current))

            if len(pending) == 0:
                return Ok(topo_sorted)
        else:
            current += 1

        if current == len(pending):
            current = 0

    return Ok(topo_sorted)


# DAG querying
def downstream(dag, node):
    return upstream(reverse_dict_inclusive(dag), node)


def upstream(dag, node):
    to_include = list()
    queue = dag[node]
    while len(queue) > 0:
        current = queue.pop(0)
        if current not in to_include:
            to_include.append(current)
            queue.extend(dag[current])

    return Ok(to_include)


def query(dag, query=list()):
    result = topological_sort(dag)
    if result.is_err:
        return result
    elif len(query) == 0:
        return Ok(result.value)
    else:
        topo_sort = result.value

    query = sorted(query, key=lambda x: 0 if x["operation"] == "include" else 1)
    if query[0]["operation"] == "include":
        to_include = set()
    else:
        to_include = set(topo_sort)
    query_cache = dict()

    for operand in query:
        key = (operand["task"], operand["upstream"], operand["downstream"])
        if key in query_cache:
            tasks = query_cache[key]
        else:
            tasks = [key[0]]
            if key[1]:
                result = upstream(dag, key[0])
                if result.is_err:
                    return result
                tasks.extend(result.value)
            if key[2]:
                result = downstream(dag, key[0])
                if result.is_err:
                    return result
                tasks.extend(result.value)
            query_cache[key] = tasks

        if operand["operation"] == "include":
            to_include = to_include.union(tasks)
        else:
            to_include -= set(tasks)

    return Ok([n for n in topo_sort if n in to_include])



import random
from uuid import uuid4

from sayn import task


@task(outputs=["logs_arenas", "logs_tournaments", "logs_battles", "logs_fighters"])
def load_data(context, warehouse):
    fighters = ["Son Goku", "John", "Lucy", "Dr. x", "Carlos", "Dr. y?"]
    arenas = ["Earth Canyon", "World Edge", "Namek", "Volcanic Crater", "Underwater"]
    tournaments = ["World Championships", "Tenka-ichi Budokai", "King of the Mountain"]

    context.set_run_steps(
        [
            "Generate Dimensions",
            "Generate Battles",
            "Load fighters",
            "Load arenas",
            "Load tournaments",
            "Load battles",
        ]
    )

    with context.step("Generate Dimensions"):
        # Add ids to the dimensions
        fighters = [
            {"fighter_id": str(uuid4()), "fighter_name": val}
            for id, val in enumerate(fighters)
        ]
        arenas = [
            {"arena_id": str(uuid4()), "arena_name": val}
            for id, val in enumerate(arenas)
        ]
        tournaments = [
            {"tournament_id": str(uuid4()), "tournament_name": val}
            for id, val in enumerate(tournaments)
        ]

    with context.step("Generate Battles"):
        battles = list()
        for tournament in tournaments:
            tournament_id = tournament["tournament_id"]
            # Randomly select a number of battles to generate for each tournament
            n_battles = random.choice([10, 20, 30])

            for _ in range(n_battles):
                battle_id = str(uuid4())

                # Randomly choose fighters and arena
                fighter1_id = random.choice(fighters)["fighter_id"]
                fighter2_id = random.choice(
                    [f for f in fighters if f["fighter_id"] != fighter1_id]
                )["fighter_id"]
                arena_id = random.choice(arenas)["arena_id"]

                # Pick a winner
                winner_id = fighter1_id if random.uniform(0, 1) <= 0.5 else fighter2_id

                battles.append(
                    {
                        "event_id": str(uuid4()),
                        "tournament_id": tournament_id,
                        "battle_id": battle_id,
                        "arena_id": arena_id,
                        "fighter1_id": fighter1_id,
                        "fighter2_id": fighter2_id,
                        "winner_id": winner_id,
                    }
                )

    data_to_load = {
        "fighters": fighters,
        "arenas": arenas,
        "tournaments": tournaments,
        "battles": battles,
    }

    # Load logs
    for log_type, log_data in data_to_load.items():
        with context.step(f"Load {log_type}"):
            warehouse.load_data(f"logs_{log_type}", log_data, replace=True)



from copy import deepcopy
from typing import Any, Dict, Optional, Set

from ..database.unknown import UnknownDb

from ..core.errors import Err, Exc, Ok, Result
from ..utils.misc import map_nested

from .task import Task, TaskStatus

# Properties from a task dictionary that won't be part of the task
# configuration as they are standard SAYN task properties
_excluded_properties = (
    "name",
    "type",
    "tags",
    "group",
    "parents",
    "sources",
    "outputs",
    "parameters",
    "class",
    "preset",
    "on_fail",
    "module",
)


class TaskWrapper:
    """Task wrapper managing the execution of tasks.

    This class wraps a task runner (sayn.tasks.Task) and manages the lifetime of said task.
    In a SAYN execution there will be 1 TaskWrapper per task in the project, whether the task
    is to be executed or ignored based on the task query (`-t` and `-x` arguments to `sayn`).

    Properties:
      name (str): the name of the task
      group (str): the name of the task group file where the task was defined
      tags (List[str]): list of tags declared for the task
      parents (List[sayn.task_wrapper.TaskWrapper]): the list of parents to the current task
      project_parameters (Dict[str, Any]): parameters defined at project level
      task_parameters (Dict[str, Any]): parameters defined at task level
      in_query (bool): whether the task is selected for execution based on the task query
      runner (Task): the object that will do the actual work
      status (TaskStatus): the current status of the task
    """

    name: str
    group: str
    tags: Set[str]
    parent_names: Set[str]
    sources_yaml: Set[Any]
    outputs_yaml: Set[Any]
    used_connections: Set[str]
    project_parameters: Dict[str, Any]
    task_parameters: Dict[str, Any]
    in_query: bool = False
    runner: Optional[Task]
    status: TaskStatus = TaskStatus.UNKNOWN

    def __init__(
        self,
        group: str,
        name: str,
        task_type: str,
        on_fail: str,
        parent_names: Set[str],
        sources: Set[str],
        outputs: Set[str],
        tags: Set[str],
        tracker,
        task_class,
        connections,
        default_db,
        run_arguments,
        compiler,
        db_object_compiler,
    ):

        self.tags = set(tags or set())
        self.parent_names = set(parent_names or set())
        self.parents = list()
        self.sources_yaml = set(sources or set())
        self.sources = set()
        self.outputs_yaml = set(outputs or set())
        self.outputs = set()
        self.sources_from_prod = set()
        self.used_connections = set()
        self.tracker = tracker
        self.runner = None

        self.name = name
        self.group = group
        self.task_type = task_type
        self.on_fail = on_fail or "skip"

        self.default_db = default_db
        self.connections = dict(connections)
        self.db_object_compiler = db_object_compiler

        self.task_class = task_class

        if self.task_class is None:
            self.status = TaskStatus.FAILED
        else:
            self.status = TaskStatus.CONFIGURING

            self.compiler = compiler.get_task_compiler(group=self.group, name=self.name)

            self.run_arguments = {
                "debug": run_arguments.debug,
                "full_load": run_arguments.full_load,
                "start_dt": run_arguments.start_dt,
                "end_dt": run_arguments.end_dt,
                "dates_specified": run_arguments.dates_specified,
                "command": run_arguments.command.value,
                "is_prod": run_arguments.is_prod,
                "folders": {
                    "python": run_arguments.folders.python,
                    "sql": run_arguments.folders.sql,
                    "compile": run_arguments.folders.compile,
                    "logs": run_arguments.folders.logs,
                    "tests": run_arguments.folders.tests,
                },
            }

    def config(
        self,
        task_config: Dict[str, Any],
        project_parameters: Dict[str, Any],
        task_parameters: Dict[str, Any],
    ):
        # Check the parents are in a good state
        result = self.check_skip()
        if result.is_err or result.value == TaskStatus.SKIPPED:
            return result

        try:
            self.set_parameters(project_parameters, task_parameters)
        except Exception as exc:
            self.status = TaskStatus.FAILED
            return Exc(exc, where="set_task_parameters")

        # for decorator tasks
        if hasattr(self.task_class, "func_arguments"):
            for arg in self.task_class.func_arguments:
                if arg in self.connections:
                    self.used_connections.add(arg)

        try:
            runner = self.task_class(
                self.name,
                self.group,
                self.tracker,
                self.run_arguments,
                self.task_parameters,
                self.project_parameters,
                self.default_db,
                self.connections,
                self.compiler,
                self.src,
                self.out,
            )
        except Exception as exc:
            self.status = TaskStatus.FAILED
            return Exc(exc, where="create_task_runner")

        # # compile Outputs YAML and Sources YAML
        self.outputs_yaml = [self.compiler.compile(o) for o in self.outputs_yaml]

        self.sources_yaml = [self.compiler.compile(o) for o in self.sources_yaml]
        # convert sources and outputs from strings in yaml
        # Now we can compile the other properties
        runner_config = {
            k: v for k, v in task_config.items() if k not in _excluded_properties
        }

        try:
            runner_config = map_nested(
                runner_config,
                lambda x: self.compiler.compile(x) if isinstance(x, str) else x,
            )
        except Exception as exc:
            return Exc(exc, where="compile_task_properties")

        self.runner = runner

        try:
            result = self.runner.config(**runner_config)
            if result is not None and result.is_err:
                self.status = TaskStatus.FAILED
                return result
        except Exception as exc:
            self.status = TaskStatus.FAILED
            return Exc(exc, where="task_config")

        # The config stage can produce changes to: tags, parents, sources, outputs and on_fail
        if hasattr(runner, "_config_input"):
            for source in runner._config_input["sources"]:
                self.src(source)

            for output in runner._config_input["outputs"]:
                self.out(output)

            for parent in runner._config_input["parents"]:
                self.parent_names.add(parent)

            for tag in runner._config_input["tags"]:
                self.tags.add(tag)

            if runner._config_input.get("on_fail") is not None:
                self.on_fail = runner._config_input["on_fail"]

            # A bit of cleaning on the user
            del runner._config_input

        # TODO relying on the task object having a certain property is not a solid method. Need to change it
        if "_target_db" in runner.__dict__:
            target_connection = self.connections[runner._target_db]
        else:
            target_connection = self.connections[self.default_db]

        self.outputs.update(
            {
                self.db_object_compiler.from_string(o, connection=target_connection)
                for o in self.outputs_yaml
            }
        )

        if "_source_db" in runner.__dict__:
            source_connection = self.connections[runner._source_db]
        elif "_target_db" in runner.__dict__:
            source_connection = self.connections[runner._target_db]
        else:
            source_connection = self.connections[self.default_db]

        self.sources.update(
            {
                self.db_object_compiler.from_string(o, connection=source_connection)
                for o in self.sources_yaml
            }
        )

        self.status = TaskStatus.READY_FOR_SETUP

        self.verify_connections()

        return Ok()

    def set_parameters(
        self, project_parameters: Dict[str, Any], task_parameters: Dict[str, Any]
    ):
        # Process parameters
        # The project parameters go as they come
        self.project_parameters = deepcopy(project_parameters or dict())

        # Compile nested dictionary of parameters
        task_parameters = task_parameters or dict()
        task_parameters = map_nested(
            task_parameters,
            lambda x: self.compiler.compile(x) if isinstance(x, str) else x,
        )
        self.task_parameters = task_parameters

        # Add the task paramters to the jinja environment
        self.compiler.update_globals(**task_parameters)

    def check_skip(self):
        if self.run_arguments["command"] != "test":
            failed_parents = {
                p.name: p.status
                for p in self.parents
                if (
                    p.status in (TaskStatus.SETUP_FAILED, TaskStatus.FAILED)
                    and p.on_fail != "no_skip"
                )
                or p.status == TaskStatus.SKIPPED
            }
        else:
            failed_parents = {}

        if len(failed_parents) > 0:
            self.status = TaskStatus.SKIPPED

            if self.in_query:
                return Err("task", "parent_errors", failed_parents=failed_parents)
            else:
                return Ok(self.status)

        elif self.status == TaskStatus.SETUP_FAILED:
            return Err("task", "setup_error", status=self.status)
        elif "fail" in self.status.value:
            return Err("task", "task_error", status=self.status)
        else:
            return Ok(self.status)

    def setup(self, in_query, sources_from_prod):
        # Check the parents are in a good state
        result = self.check_skip()
        if result.is_err or result.value == TaskStatus.SKIPPED:
            return result

        if self.runner is None:
            return Ok()

        self.in_query = in_query
        self.status = TaskStatus.SETTING_UP

        if not in_query:
            self.status = TaskStatus.NOT_IN_QUERY
            return Ok()
        else:
            # Clear all the dummy connections
            self.runner.connections = {
                n: None if isinstance(o, UnknownDb) else o
                for n, o in self.runner.connections.items()
            }

            needs_recompile = False
            for s in self.sources:
                if s in sources_from_prod:
                    needs_recompile = True

            self.runner._needs_recompile = needs_recompile

            self.sources_from_prod = sources_from_prod

            # Run the setup stage for the runner and return the results
            try:
                result = self.runner.setup()
            except Exception as exc:
                result = Exc(exc)

            finally:
                if result is None:
                    self.status = TaskStatus.READY
                    return Ok()
                elif not isinstance(result, Result):
                    self.status = TaskStatus.SETUP_FAILED
                    self.tracker.current_step = None
                    return Err("task_result", "missing_result_error", result=result)
                elif result.is_err:
                    self.status = TaskStatus.SETUP_FAILED
                    self.tracker.current_step = None
                    return result
                else:
                    self.status = TaskStatus.READY
                    return Ok()

    def should_run(self):
        return self.status == TaskStatus.NOT_IN_QUERY

    def run(self):
        return self.execute_task("run")

    def compile(self):
        return self.execute_task("compile")

    def test(self):
        return self.execute_task("test")

    def execute_task(self, command):
        result = self.check_skip()
        if result.is_err or result.value == TaskStatus.SKIPPED:
            return result

        if self.runner is None:
            return Ok()

        if not self.in_query:
            # TODO review this as it should never happend if running sayn cli
            self.status = TaskStatus.NOT_IN_QUERY
            return Err("execution", "task_not_in_query")
        elif self.status not in (TaskStatus.SETTING_UP, TaskStatus.READY):
            return Err("execution", "setup_error", status=self.status)
        else:
            try:
                if command == "run":
                    result = self.runner.run()
                elif command == "compile":
                    result = self.runner.compile()
                else:
                    result = self.runner.test()

                if result is None:
                    result = Ok()
                    self.status = TaskStatus.SUCCEEDED
                elif result.is_ok:
                    self.status = TaskStatus.SUCCEEDED
                else:
                    self.status = TaskStatus.FAILED
            except Exception as e:
                self.status = TaskStatus.FAILED
                result = Exc(e)

            return result

    def set_parents(self, all_tasks, output_to_task):
        for parent_name in self.parent_names:
            if parent_name not in all_tasks:
                return Err("dag", "missing_parents", missing={self.name: [parent_name]})
            self.parents.append(all_tasks[parent_name])
            self.parent_names.add(parent_name)

        missing = set()
        for source in self.sources:
            if source not in output_to_task:
                if source.connection_name == self.default_db:
                    # We only consider a missing parent if the
                    # source is in the default_db
                    missing.add(source)
            else:
                for task_name in output_to_task[source]:
                    if task_name not in self.parent_names:
                        self.parents.append(all_tasks[task_name])
                        self.parent_names.add(task_name)

        # TODO send some message when a table is source
        if len(missing) > 0:
            tables = ", ".join([f"{t.raw}" for t in missing])
            self.tracker.warning(
                f'No task creates table(s) "{tables}" referenced by task "{self.name}"'
            )

        return Ok()

    def src(self, obj, connection=None, level=None):
        obj = self.db_object_compiler.from_string(
            obj, connection=connection, level=level
        )
        if self.status == TaskStatus.CONFIGURING:
            # During configuration we add to the list and use values based on settings
            self.used_connections.add(obj.connection_name)
            self.sources.add(obj)

        return self.db_object_compiler.src_value(obj)

    def out(self, obj, connection=None, level=None):
        obj = self.db_object_compiler.from_string(
            obj, connection=connection, level=level
        )
        if self.status == TaskStatus.CONFIGURING:
            self.used_connections.add(obj.connection_name)
            self.outputs.add(obj)

        return self.db_object_compiler.out_value(obj)

    def verify_connections(self):
        if hasattr(self.runner, "_target_db"):
            self.used_connections.add(self.runner._target_db)

        if hasattr(self.runner, "_source_db"):
            self.used_connections.add(self.runner._source_db)
        return Ok()

    def has_tests(self):
        if self.runner is not None and self.runner._has_tests:
            return True

        return False
